"""
Bot sources module. Defines classes of distinct bot sources that behave
differently at the various steps of a factory functionality

Constants

    SUPPORTED_SOURCES:  Dictionary of <source_identifier>: <source_subclass>
"""
import importlib
import os
import re
import shutil
import subprocess
import hashlib
import sys
from pathlib import Path

try:
    import git
except ImportError:
    # ImportError: Most likely git is not in the expected place.  Not an issue
    #   unless the user wants to use bot_sources.GitBotSource
    pass
from click import BadParameter
from colorama import Fore

from pyqalx.core.errors import QalxFactoryValidationError, QalxFactoryBuildError
from pyqalx.core.utils import _msg_user


class BotSource:
    """
    Base class for source related functionality about a specific bot

    Instance attributes

        bot_name:       The bot name
        bot:            The entire bot definition from the plan
        build_path:     The bot build path
        md5_hash:       The md5_hash of the bot code
    """

    FORE = Fore.MAGENTA

    def __init__(self, bot_name, bot, build_path_dir, log, *args, **kwargs):
        self.bot_name = bot_name
        self.bot = bot
        self.source_dict = bot["source"]
        self.build_path = os.path.join(build_path_dir, bot_name)
        self.md5_hash = None
        self.log = log

    @staticmethod
    def get_source(bot):
        """
        Static method that accepts a bot definition and tries to identify
        the source type and return the appropriate BotSource subclass if found

        :param bot: The bot definition from the plan
        :return: The BotSource subclass if found

        :raises KeyError:     If no bot source subclass is identified
        """
        source = bot["source"]
        for key in source:
            bot_source_class = SUPPORTED_SOURCES.get(key.lower())
            if bot_source_class:
                return bot_source_class
        raise KeyError(f"Unknown bot source `{source}`")

    def _msg_user(self, msg, **kwargs):
        """
        Helper method that prints and also logs a debug message if a log
        instance attribute exists
        """
        _msg_user(
            title=self.bot_name, fore=self.FORE, msg=msg, log=self.log, **kwargs
        )

    def zip_build_path(self):
        """
        Zips the contents of the build_path. The name of the file is:
        "<name_of_build_path>.zip" and it is located in the parent folder of
        build_path

        :return: The path to the zip file created
        """
        return shutil.make_archive(self.build_path, "zip", self.build_path)

    @staticmethod
    def get_single_hash(file_path):
        """
        The an md5 hash for a single file. Bytes are read in chunks for a given
        file

        :param file_path: The path to the file to get an md5 hash for
        :return: The md5 hash
        """
        hash_md5 = hashlib.md5()
        with open(file_path, "rb") as f:
            # Reading the file in chunks to avoid reading the whole file in
            # memory: https://stackoverflow.com/a/3431838
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    def get_hash(self):
        """
        Calculates the md5 hash of the contents of the build path, the value of
        which is stored under the md5_hash instance attribute
        """
        self.md5_hash = self._get_hash()

    def _get_hash(self):
        """
        Hook for calculating the md5 hash for each bot source type
        """
        self._msg_user("Calculating hash")
        hashes = []
        for root, _, files in os.walk(self.build_path):
            for file in files:
                fp = os.path.join(root, file)
                hashes.append(self.get_single_hash(fp))
        hasher = hashlib.md5()
        for hash_value in sorted(hashes):
            hasher.update(hash_value.encode("utf-8"))
        self._msg_user("Finished calculating hash")
        return hasher.hexdigest()

    def validate(self):
        """
        Hook for providing validation on a BotSource. Called by `download_bot`
        after the download has completed
        """
        self._validate_bot_path()

    def _validate_bot_path(self):
        """
        Validates that the bot_path is importable.  Adds the build_path to the
        path temporarily so that we can attempt the import.
        """
        from qalxcli.cli_types import BOT_IMPORT

        sys.path.append(self.build_path)

        def _remove_from_path():
            sys.path.remove(self.build_path)

        try:
            BOT_IMPORT(self.bot["bot_path"])
        except BadParameter as exc:
            # Checking if the exception is related to an ImportError through the
            # exception message. In such a case, we do not want to error, but
            # only warn the user that the build might fail later.
            if "No module named" in str(exc):
                module_regex = re.compile(".*No module named '(.*)'.*")
                module_name = module_regex.match(str(exc))[1]
                msg = (
                    f"'{module_name}' could not be imported during bot download"
                    ". Your factory build may fail. We have no way of knowing"
                )
                self._msg_user(msg)
            else:
                raise QalxFactoryValidationError(
                    f"Error validating `bot_path` on `{self.bot_name}`: {exc}"
                )
        finally:
            _remove_from_path()

    def _download_bot(self):
        """Hook for downloading a bot for a specific BotSource"""
        raise NotImplementedError("Must be implemented by subclass")

    def download_bot(self):
        bot = self._download_bot()
        self.validate()
        return bot

    def install_bot(self):
        """
        Install the bot code. This is needed for the bot code package and the
        dependencies to be correctly installed and configured
        """
        self._msg_user("Installing...")
        self._install_bot()
        self._msg_user("Finished Installing...")

    def _install_paths(self):
        """
        Starting at the same directory that the bot file resides in (the bottom
        of the tree), walk the path upwards until we get to the root of the
        build path.  Each of these directories `could` contain a setup.py,
        pyproject.toml or requirements.txt.
        :return: A `path.Path` instance where the install file `could` be
        """
        # Ensure we start at the lowest directory possible.
        # i.e. if user supplies the following:
        #       source:
        #         path: /path/to/factories/myfactory/my_bot_3_dir
        #       bot_path: deeply.nested.structure.my_bot_file_3:my_bot_3
        #
        # we start at: /the/build/path/my_bot_3/deeply/nested/structure
        # then walk upwards up to and including /the/build/path/my_bot_3/
        bot_file_dir = (
            self.bot["bot_path"].rpartition(".")[0].replace(".", os.path.sep)
        )
        install_path = Path(self.build_path, bot_file_dir)
        while install_path != Path(self.build_path).parent.absolute():
            # Walk up the tree until we're at the root of the build path
            yield install_path
            install_path = Path(install_path).parent.absolute()

    def _install_bot(self):
        """
        Hook for installing the bot code using pip and code that has been
        downloaded in the build path.  Walks from the bottom of the tree upwards
        until an install file has been found.  Once a file has been found and
        install has completed successfully we stop looking for further install
        files
        """
        for install_path in self._install_paths():
            # First try to install from `pyproject.toml` or `setup.py`.
            self._msg_user(
                "Attempting to install `setup.py` or `pyproject.toml` "
                f"from `{install_path}`"
            )
            output = subprocess.run(
                [shutil.which("pip"), "install", ".", "--upgrade"],
                cwd=install_path,
                check=False,
            )
            if output.returncode == 0:
                return
            self._msg_user(
                "Unable to install dependencies from `setup.py` "
                f"or `pyproject.toml` from `{install_path}`.  "
                f"Attempting to install from requirements file in `{install_path}`."
            )

            # The fallback option is to use the user defined requirements.txt file
            # if specified - failing that attempt to use a `requirements.txt` file
            # on `self.bot_path`
            requirements_path_key = "requirements_path"
            requirements_path = self.bot.get(
                requirements_path_key, "requirements.txt"
            )
            if requirements_path_key in self.bot and not os.path.exists(
                requirements_path
            ):
                # If the user has specified requirements we should check the
                # file exists
                msg = f"{requirements_path} does not exist."
                raise QalxFactoryBuildError(msg)
            output = subprocess.run(
                [
                    shutil.which("pip"),
                    "install",
                    "-r",
                    requirements_path,
                    "--upgrade",
                ],
                cwd=install_path,
                check=False,
            )

            if output.returncode == 0:
                return
            if os.path.exists(requirements_path):
                # The requirements path existed but there was an error installing.
                # An exception is raised so the user can go and fix it
                raise QalxFactoryBuildError(
                    f"An error occurred installing"
                    f" dependencies "
                    f"from {requirements_path}."
                )
            # There was no file found to install dependencies from in this
            # directory.  Continue walking the path upwards
            self._msg_user(
                f"No files found to install dependencies from in {install_path}."
            )
        # There was no file found to install dependencies from in any of the
        # searched directories.  This may be intentional but we have no way of
        # knowing.  Inform the user.
        self._msg_user(
            "No files found to install dependencies from."
            "  Dependencies not installed."
        )


class PathBotSource(BotSource):
    """
    Bot source class for code provide from a local path

    Instance attributes
        path:   The local path to the bot code
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.path = self.source_dict.get("path")

    def _download_bot(self):
        """
        Copies the local bot code to build_path
        """
        msg = f"Copying contents from `{self.path}` to `{self.build_path}`"
        self._msg_user(msg)
        shutil.copytree(self.path, self.build_path)


class GitBotSource(BotSource):
    """
    Bot source class for code provided from a url of a git repository

    Instance attributes
        url:   The url to the git repository with the bot code
    """

    def __init__(self, *args, **kwargs):
        # Ensure that the git is importable - required for GitBotSource to work
        importlib.import_module("git")

        super().__init__(*args, **kwargs)
        self.url = self.source_dict.pop("url")

    def _download_bot(self):
        """
        Downloads the bot code to build_path
        """
        msg = f"Cloning repository from `{self.url}` to `{self.build_path}`"
        self._msg_user(msg)
        git.Repo.clone_from(self.url, self.build_path, **self.source_dict)


class PyPIBotSource(BotSource):
    """
    Bot source class for code provided from a pypi repository

    Class attributes

        PYPI:   The key for pypi on the bot source
    """

    PYPI = "pypi"

    @staticmethod
    def install(source_dict):
        """
        method that allows easier installation of pypi sources for remote
        sectors
        :param source_dict:The source dict
        """
        pypi = source_dict.get(PyPIBotSource.PYPI)
        version = source_dict.get("version")
        if version is not None:
            pypi += f"=={version}"

        output = subprocess.run(
            [shutil.which("pip"), "install", pypi, "--upgrade"], check=False
        )
        if output.returncode != 0:
            raise QalxFactoryBuildError(f"Failed to install {pypi}")

    def zip_build_path(self):
        """
        pypi sources don't get downloaded - so there is nothing to zip
        """

    def _install_bot(self):
        """
        Pypi bots get installed directly from source
        """
        PyPIBotSource.install(self.source_dict)

    def _validate_bot_path(self):
        """
        bot_path cannot be validated for pypi as the package won't
        be installed so there is nothing that can be imported.
        """

    def _download_bot(self):
        """
        `pypi` sources don't get downloaded.  This is because, if the associated
        bot is for a remote sector the OS could be different - meaning that
        what is downloaded may not work on the remote sector.  Instead, these
        are just installed when `build` is called. An item will still be created
        but it won't have any data on it.
        """
        self._msg_user("Found pypi source.  Not downloading")

    def _get_hash(self):
        """
        Because no file is downloaded there is nothing to calculate a hash on.
        The API cannot be queried as the hashes are calculated for a specific
        file type (.tar.gz, .whl etc) so there is no way to be consistent.
        """
        self._msg_user("Found pypi source.  Not generating hash")


class QalxItemBotSource(BotSource):
    """
    Source class for Items that contain Bot code as a zip file.  Used by
    remote sectors to download and extract the bot code.

    Instance attributes

        bot_name:       The bot name
        bot:            The entire bot definition from the plan
        build_path:     The bot build path
        bot_code_item:  The ~pyqalx.Item instance that contains the zipped bot
                        code
    """

    def __init__(self, bot_code_item, *args, **kwargs):
        super(QalxItemBotSource, self).__init__(*args, **kwargs)
        self.bot_code_item = bot_code_item

    def _install_bot(self):
        if self.source_dict.get(PyPIBotSource.PYPI):
            # ensure pypi bots get installed in a specific way
            PyPIBotSource.install(source_dict=self.source_dict)
        else:
            super(QalxItemBotSource, self)._install_bot()

    def _download_bot(self):
        """
        Downloads the bot code from the Item and saves it in the appropriate
        build path for this bot.  Unzips the zip file.  Installs the bot
        """
        msg = f"Downloading bot code from API to `{self.build_path}`"
        self._msg_user(msg)
        if not os.path.exists(self.build_path):
            # The <bot_name> sub directory probably doesn't exist at bootstrap
            # time.  Ensure it gets created
            os.mkdir(self.build_path)

        if not self.source_dict.get(PyPIBotSource.PYPI):
            # Save the zipfile to disk.  Pypi bots won't have a file
            filename = self.bot_code_item.save_file_to_disk(self.build_path)
            # Unzip the zipped bot code into the build path.  This path will get
            # added to the PYTHONPATH at bot start time.
            shutil.unpack_archive(
                filename=filename, extract_dir=self.build_path
            )
        # Bots on remote instances should always be installed into the
        # environment immediately after downloading
        self.install_bot()
        # Return the bot_name as this is required for starting the bot
        return self.bot_name


SUPPORTED_SOURCES = {
    "url": GitBotSource,
    "path": PathBotSource,
    "pypi": PyPIBotSource,
}



from pyqalx.core.adapters.adapter import (
    QalxSignalAdapter,
    QalxUpdateStatusAdapter,
    QalxUpdateStateAdapter,
)
from pyqalx.core.entities.worker import Worker
from pyqalx.core.enums import ValidStates
from pyqalx.core.signals import QalxWorkerSignal


class QalxWorker(
    QalxSignalAdapter, QalxUpdateStatusAdapter, QalxUpdateStateAdapter
):
    """
    Provides an interface for accessing the API for
    :class:`~pyqalx.core.entities.worker.Worker` entities
    """

    _entity_class = Worker
    signal_class = QalxWorkerSignal

    def _process_api_request(self, *args, **kwargs):
        """
        Ensures that the `bot_entity` kwarg doesn't get sent to the API. This
        variable will have been passed down through kwargs from any endpoint
        method and is required for building the urls for a worker
        """
        # `bot_entity` may not be in kwargs anymore if we are calling
        # `_process_api_request` twice (i.e. when getting a list of workers)
        kwargs.pop("bot_entity", None)
        return super(QalxWorker, self)._process_api_request(*args, **kwargs)

    def list_endpoint(self, *args, bot_entity, **kwargs):
        """
        Builds the list_endpoint for workers.  This requires the bot_entity
        which will get passed down from the calling method via kwargs.
        :param bot_entity:An instance of Bot
        :type bot_entity:~entities.bot.Bot
        :return:The Worker list endpoint
        """
        bot_endpoint = self.session.bot.detail_endpoint(guid=bot_entity["guid"])
        return f"{bot_endpoint}/{self.entity_class.entity_type}"

    def detail_endpoint(self, guid, *args, bot_entity, **kwargs):
        """
        Builds the detail_endpoint for workers.  This requires the bot_entity
        which will get passed down from the calling method via kwargs.

        :param guid: The guid of the Worker
        :type guid: `uuid`
        :param bot_entity: An instance of Bot
        :type bot_entity: ~pyqalx.core.entities.bot.Bot
        :return: The Worker list endpoint
        """
        bot_endpoint = self.list_endpoint(bot_entity=bot_entity)
        return f"{bot_endpoint}/{guid}"

    def get(self, guid, bot_entity, *args, **kwargs):
        """
        Gets an individual worker.

        :param guid: The guid of the Worker
        :type guid: `uuid`
        :param bot_entity: An instance of Bot
        :type bot_entity: ~pyqalx.core.entities.bot.Bot
        :return: :class:`~pyqalx.core.entities.worker.Worker`
        """
        return super(QalxWorker, self).get(
            guid, bot_entity=bot_entity, *args, **kwargs
        )

    def get_signal(self, entity, bot_entity, *args, **kwargs):
        """
        Gets the signal for a specific entity

        :param entity: An instance of Worker
        :type entity: ~pyqalx.core.entities.worker.Worker
        :param bot_entity: An instance of Bot
        :type bot_entity: ~pyqalx.core.entities.bot.Bot
        :return: :class:`~pyqalx.core.signals.QalxWorkerSignal`
        """
        return super(QalxWorker, self).get_signal(
            entity=entity, bot_entity=bot_entity, *args, **kwargs
        )

    def terminate(self, entity, bot_entity, *args, **kwargs):
        """
        Terminates a specific entity

        :param entity: An instance of Worker
        :type entity: ~pyqalx.core.entities.worker.Worker
        :param bot_entity: An instance of Bot
        :type bot_entity: ~pyqalx.core.entities.bot.Bot
        """
        self.update_state(entity, ValidStates.TERMINATED, bot_entity=bot_entity)
        return super(QalxWorker, self).terminate(
            entity=entity, bot_entity=bot_entity, *args, **kwargs
        )

    def _stop_or_resume(self, entity, stop, bot_entity, **kwargs):
        """
        Stops or resumes a specific entity

        :param entity: An instance of Worker
        :type entity: ~pyqalx.core.entities.worker.Worker
        :param stop: Whether we should stop or resume
        :type stop: bool
        :param bot_entity: An instance of Bot
        :type bot_entity: ~pyqalx.core.entities.bot.Bot
        """
        # Update to `idle` on a resume - if there are extant jobs the state
        # will be updated to active at the next process function
        state = ValidStates.STOPPED if stop else ValidStates.IDLE
        self.update_state(entity, state, bot_entity=bot_entity)
        return super(QalxWorker, self)._stop_or_resume(
            entity=entity, stop=stop, bot_entity=bot_entity, **kwargs
        )

    def reload(self, entity, bot_entity, **kwargs):
        """
        Reloads the current entity from the API

        :param entity: An instance of Worker
        :type entity: ~pyqalx.core.entities.worker.Worker
        :param bot_entity: An instance of Bot
        :type bot_entity: ~pyqalx.core.entities.bot.Bot
        :return: :class:`~pyqalx.core.entities.worker.Worker`
        """
        return super(QalxWorker, self).reload(
            entity=entity, bot_entity=bot_entity
        )

    def update_status(self, entity, bot_entity, status):
        """
        Updates the workers status

        :param entity: An instance of Worker
        :type entity: ~pyqalx.core.entities.worker.Worker
        :param bot_entity: An instance of Bot
        :type bot_entity: ~pyqalx.core.entities.bot.Bot
        :param status: The status to update to
        :type status: str
        """
        return super(QalxWorker, self).update_status(
            entity=entity, bot_entity=bot_entity, status=status
        )

    def stop(self, entity, bot_entity, **kwargs):
        """
        Sends a stop signal to the worker

        :param entity: An instance of Worker
        :type entity: ~pyqalx.core.entities.worker.Worker
        :param bot_entity: An instance of Bot
        :type bot_entity: ~pyqalx.core.entities.bot.Bot
        """
        return super(QalxWorker, self).stop(
            entity=entity, bot_entity=bot_entity, **kwargs
        )

    def resume(self, entity, bot_entity, **kwargs):
        """
        Sends a resume signal to the worker

        :param entity: An instance of Worker
        :type entity: ~pyqalx.core.entities.worker.Worker
        :param bot_entity: An instance of Bot
        :type bot_entity: ~pyqalx.core.entities.bot.Bot
        """
        return super(QalxWorker, self).resume(
            entity=entity, bot_entity=bot_entity, **kwargs
        )

    def save(self, entity, bot_entity, *args, **kwargs):
        """
        Saves the Worker instance

        :param entity: An instance of Worker
        :type entity: ~pyqalx.core.entities.worker.Worker
        :param bot_entity: An instance of Bot
        :type bot_entity: ~pyqalx.core.entities.bot.Bot
        :return: :class:`~pyqalx.core.entities.worker.Worker`
        """
        return super(QalxWorker, self).save(
            entity=entity, bot_entity=bot_entity, *args, **kwargs
        )



from pyqalx.core.adapters.adapter import QalxUnpackableAdapter
from pyqalx.core.entities import Group


class QalxGroup(QalxUnpackableAdapter):
    """
    Provides an interface for accessing the API for
    :class:`~pyqalx.core.entities.group.Group` entities
    """

    _entity_class = Group
    child_entity_class = Group.child_entity_class

    def add(self, sets, meta=None, **kwargs):
        """
        When adding a `Group` ensure that the sets posted to the api are in
        the format {<key>: :class:`~pyqalx.core.entities.set.Set`}

        :param sets: A dictionary of Sets to create on the group
        :type sets: dict
        :param meta: A dictionary of metadata to store
        :type meta: dict
        :return: A newly created `Group` instance
        """
        return super(QalxGroup, self).add(sets=sets, meta=meta, **kwargs)

    def get(self, guid, child_fields=None, unpack=False, *args, **kwargs):
        """
        Gets the entity for the given `guid` and unpacks it if specified.
        Provide `child_fields` to restrict the fields returned on unpacked
        children


        :param guid: The `guid` of the entity to get
        :type guid: str
        :param child_fields: A list of fields that should be returned from
                             child entities
        :type child_fields: list
        :param unpack: Should the child entities be unpacked. Defaults to False
        :type unpack: bool
        :return: An unpacked entity
        """
        return super(QalxGroup, self).get(
            guid, child_fields=child_fields, unpack=unpack, *args, **kwargs
        )

    def reload(self, entity, unpack=False, *args, **kwargs):
        """
        Reloads the current entity from the API.

        :param entity: An instance of `self.entity_class`
        :param unpack: Should the child entities by unpacked?  Defaults to False
        :type unpack: bool
        :return: A refreshed instance of `self.entity_class`
        """
        return super(QalxGroup, self).reload(
            entity, unpack=unpack, *args, **kwargs
        )

    def find(
        self,
        query=None,
        sort=None,
        skip=0,
        limit=25,
        many=True,
        child_fields=None,
        include_session_tags=True,
        *args,
        **kwargs
    ):
        """
        Return multiple packed entities from the API

        :param query: The optional Mongo query to find entities
        :type query: dict
        :param sort: The keys to sort by
        :type sort: list
        :param skip: The number of results to skip (offset) by
        :type skip: int
        :param limit: How many results should the response be limited to
        :type limit: int
        :param many: Should many entities be returned or just a single one
        :type many: bool
        :param include_session_tags:    Should the tags for the session be
                                        included in the query. Default is True
        :type include_session_tags: bool
        :param child_fields: A list of fields that should be returned from
                             child entities
        :type child_fields: list
        :return: A list of entities
        """
        # extended to create better autodocs
        return super(QalxGroup, self).find(
            query=query,
            sort=sort,
            skip=skip,
            limit=limit,
            many=many,
            include_session_tags=include_session_tags,
            child_fields=child_fields,
            *args,
            **kwargs
        )

    def find_one(self, query, unpack=False, **kwargs):
        """
        Method for returning a unique entity. Will return the entity that
        matches the query

        :param query: The mongoDB query
        :type query: dict
        :param unpack: Should any child entities be automatically unpacked?
        :type unpack: bool
        :return: :class:`~pyqalx.core.entities.group.Group`
        :raises QalxMultipleEntityReturned:
        :raises QalxEntityNotFound:
        """
        return super(QalxGroup, self).find_one(query, unpack=unpack, **kwargs)

    def aggregate(self, aggregate, **kwargs):
        # extended to create better autodocs
        return super(QalxGroup, self).aggregate(aggregate=aggregate, **kwargs)

    def save(self, entity, *args, **kwargs):
        # extended to create better autodocs
        return super(QalxGroup, self).save(entity=entity, *args, **kwargs)



from colorama import Fore

from pyqalx.__version__ import __version__
from pyqalx.core.errors import QalxFactoryBuildError

import logging

from pyqalx.core.utils import _msg_user


class BaseProvider:
    """
    A default class that should handle generic functionality across all
    Providers.  This should always be subclassed for each new provider type.
    Assumes that we will be creating a single stack for each profile:region
     intersection

    Instance Attributes:
    factory:     A ~pyqalx.core.entities.factory.Factory instance
    pyqalx_configs: The configs for a User and a Bot based on the
     user_profile and bot_profile arguments on the stage.
    stage:       The stage that is being built
    stack:       The group of sectors for this stage and provider
    profile_name:  The pyqalx profile config name
    region_name:  The region that is being built for
    stack_id:   The ID of the stack.  Initialises to None and is updated when
                the stack creation call is successful
    created:  Boolean for whether the stack has finished creating
    _type: The type of the provider as a string

    :param factory:         A ~pyqalx.factories.factory.Factory instance
    :param stack:       The group of sectors for this stage and provider
    :param profile_name:  The pyqalx profile config name
    :param region_name:  The region that is being built for
    """

    _type = None
    FORE = Fore.WHITE

    def __init__(self, factory, stack, profile_name, region_name):
        self.factory = factory

        self.pyqalx_configs = {}
        self.stage = self.factory.stage
        self.stack = stack

        self.profile_name = profile_name
        self.region_name = region_name
        self.stack_id = None
        self.created = False

        if self._type is None:
            raise QalxFactoryBuildError(
                "`_type` must be defined on implementing class"
            )

    @property
    def factory_name(self):
        return self.factory.entity.name

    @classmethod
    def _send_message(cls, title, message, level, log, fore=None):
        """
        Actually send the message.  Split into a classmethod to handle
        scenarios where we don't have access to an instance but still need to
        log a message
        :param title: The title of the message
        :param message: The message to show the user/log
        :param level: The log level to log the message at
        :param log: Should the message be logged to disk or just printed to
        :param fore:Optional foreground colour.  Defaults to `cls.FORE`
        screen?
        """
        if fore is None:
            fore = cls.FORE
        title = f"{cls._type} - {title}"
        _msg_user(title=title, fore=fore, msg=message, log=log, level=level)

    def _msg_user(
        self, message, level=logging.DEBUG, should_log=True, fore=None
    ):
        """
        Prints a message to the screen for this provider instance
        and also logs it
        :param message: The message to show the user/log
        :param level: The log level to log the message at
        :param log: Should the message be logged to disk or just printed to
        screen?
        """
        log = None
        if should_log:
            # If we should log then use the log on the factory session
            log = self.factory.session.log
        sector_names = f'[{", ".join(self.stack.keys())}]'
        self._send_message(
            title=sector_names, message=message, log=log, level=level, fore=fore
        )

    @property
    def _description(self):
        return (
            f"Created by pyqalx {__version__} "
            f"for factory `{self.factory_name}`"
        )

    @staticmethod
    def validate_permissions(profile_name, region_name, log):
        """
        Validates that the calling user has permission to create the
        resources required for the stack
        """
        pass

    @staticmethod
    def default_profile_name():
        """
        The default profile to use for this provider
        :return: The default provider as a string
        """
        raise QalxFactoryBuildError("Must be implemented by subclass")

    @staticmethod
    def default_region_name(validate, log):
        """
        The default region to use for this provider
        :return: The default region as a string
        """
        raise QalxFactoryBuildError("Must be implemented by subclass")

    @property
    def ERRORED_STATUSES(self):
        """
        List of statuses that mean that the stack errored
        :return: list of statuses
        """
        raise QalxFactoryBuildError("Must be implemented by subclass")

    @property
    def CREATED_STATUS(self):
        """
        The status that means that the stack creation completed
        :return: string denoting the successful creation status
        """
        raise QalxFactoryBuildError("Must be implemented by subclass")

    def _resource_name(self, sector_name, suffix):
        """
        The name of the resource on AWS.  Appended with the factory name and
        factory guid to help ensure uniqueness
        :param sector_name: The sector that this resource belongs to
        :param suffix: The suffix to append to the name
        :return: A sanitised name for the resource
        """
        return self._sanitise_name(
            f"qalx-{self.factory_name}-{sector_name}-{suffix}"
            f"-{self.factory.entity.guid}"
        )

    @staticmethod
    def _sanitise_name(name):
        """
        Helper for sanitising the name into a cloudformation friendly format
        :param name: The name to sanitise
        :type name:str
        :return: The sanitised name.
        """
        return name.replace("_", "")

    def _stack_name(self):
        return self._sanitise_name(
            f"qalx-factory-{self.factory_name}-{self.profile_name}"
            f"-{self.factory.entity.guid}"
        )

    def _build_stack(self):
        """
        Provider specific method for building the stack template
        """
        raise QalxFactoryBuildError("Must be implemented by subclass")

    def _create_stack(self, stack):
        """
        Provider specific method for creating the stack in the provider
        :param stack: The stack template to create
        :return: The ID of the creating stack
        """
        raise QalxFactoryBuildError("Must be implemented by subclass")

    def _describe_stack(self):
        """
        Provider specific method for getting the status of the stack
        :return: The stacks status
        """
        raise QalxFactoryBuildError("Must be implemented by subclass")

    def _get_error_reason(self):
        """
        Provider specific method for getting the reasons a stack failed
        :return: list of reasons for the failed stack
        """
        raise QalxFactoryBuildError("Must be implemented by subclass")

    def _delete_stack(self, stack_id):
        """
        Provider specific method for deleting the stack
        """
        raise QalxFactoryBuildError("Must be implemented by subclass")

    def describe_stack(self):
        """
        Public interface for describing stacks
        """
        return self._describe_stack()

    def delete_stack(self, stack_id):
        """
        Public interface for deleting stacks
        """
        demolish_on_failure = self.factory._plan_debug._demolish_on_failure
        is_demolish_command = (
            self.factory.entity.status == self.factory.DELETING_STATUS
        )
        if demolish_on_failure or is_demolish_command:
            # Only demolish if the specific debug key is set on the plan OR
            # the user has run the `qalx factory-demolish` command (which updates
            # the entity status)
            self._msg_user("deleting")
            self._delete_stack(stack_id)
        else:
            # User wants to keep failed stack around.  Inform them what is
            # going on
            self._msg_user(
                self.factory._plan_debug._demolish_on_failure_msg,
                level=logging.DEBUG,
                # It's a debug message - but we want it to pop!
                fore=Fore.RED,
            )

    def create_stack(self):
        """
        Builds the stack yaml and then creates the stacks.  Calls out to
        the monitoring helper to ensure that all the stacks get created
        correctly.
        """
        # Build the stack for the specific provider
        stack = self._build_stack()
        if stack is not None:
            # Start creating the stack if there were no problems with building
            # it.
            self._msg_user("creating")
            self._create_stack(stack)



import logging
import threading
from time import sleep
import concurrent.futures

from pyqalx.factories.sectors.providers import LocalProvider


class SectorManager:
    """
    A manager that ensures that all the stacks are created
    and that if any error during creation that all existing stacks get destroyed
    """

    # An arbitrary sleep length to avoid frequent network calls
    POLL_TIME = 10

    def __init__(self, providers, plan_debug):
        self.providers = providers
        self.plan_debug = plan_debug

    def _all_completed(self):
        """
        Have all the stacks for all providers been created successfully?
        """
        return all([p.created for p in self.providers])

    @staticmethod
    def _delete_stacks(providers):
        """
        For the given list of providers will delete the stack
        :param providers: A list of Providers
        """
        for provider in providers:
            provider.delete_stack(provider.stack_id)

    def _monitor_completion(self, provider, error):
        stack_id = provider.stack_id
        stack_status = provider.describe_stack()

        while stack_status != provider.CREATED_STATUS:
            if stack_status in provider.ERRORED_STATUSES:
                # Set the error flag so we can start deleting any stacks
                # that have already created or are in the process of
                # creating.  This stack will delete itself automatically
                # due to the value of OnFailure
                error_reasons = ", ".join(provider._get_error_reason())
                msg = (
                    f"An error occured during stack creation.  "
                    f"Reason was: `{error_reasons}`"
                )
                provider._msg_user(msg, level=logging.ERROR)
                error.set()
                if isinstance(provider, LocalProvider):
                    # LocalProviders don't auto delete if they fail,
                    # so the deletion needs to be called explicitly
                    self._delete_stacks([provider])
                return False, stack_id
            if error.is_set():
                # Another stack has errored.  Therefore, delete this
                # stack before it finishes creating
                self._delete_stacks([provider])
                return provider.created, stack_id

            provider._msg_user("creating, please wait...", should_log=False)
            sleep(self.POLL_TIME)
            stack_status = provider.describe_stack()

        provider._msg_user("stack creation completed")
        provider.created = True
        return provider.created, stack_id

    def _monitor_stack_deletion(self):
        # A stack errored so any completed stacks are being completed.
        def _do_delete_check(provider):
            stack_status = provider.describe_stack()

            while (
                stack_status not in provider.DELETED_STATUSES
                and stack_status is not None
            ):
                # Wait for the stacks to delete.
                # Treat a stack that cannot be found (None status) as deleted
                stack_status = provider.describe_stack()
                provider._msg_user("deleting, please wait...", should_log=False)
                sleep(self.POLL_TIME)
            extra = "completed"
            if stack_status == provider.DELETE_FAILED:
                # better error message for the user if the stack deletion
                # fails
                extra = "failed"
            provider._msg_user(f"stack deletion {extra}")
            return True

        if self.plan_debug._demolish_on_failure:
            # Only do the delete wait check if we are actually deleting stacks,
            # otherwise we'd wait forever
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=len(self.providers)
            ) as executor:
                for provider in self.providers:
                    executor.submit(_do_delete_check, provider)

    def _monitor_stack_creation(self):
        error = threading.Event()
        results = []

        if all([p.stack_id for p in self.providers]) is False:
            # Handles the request to create a stack failing for some reason.
            # This is unlikely to happen but if a single stack can't even
            # start creation we should ensure that stacks that did start
            # creation are deleted.
            self._delete_stacks(
                filter(lambda x: x.stack_id is not None, self.providers)
            )
            self._monitor_stack_deletion()
            return
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=len(self.providers)
        ) as executor:
            for provider in self.providers:
                results.append(
                    executor.submit(self._monitor_completion, provider, error)
                )

            for future in concurrent.futures.as_completed(results):
                # Waits until all stacks complete.  If any other stacks
                # error after these have completed then we delete the completed
                # stacks.  Once all are completed then we exit

                all_completed = self._all_completed()
                while not all_completed and not error.is_set():
                    # Sit and wait until either all the stacks are completed or
                    # the error event is set (meaning that one stack failed to
                    # create)
                    all_completed = self._all_completed()
                is_created, _ = future.result()
                if is_created is True and error.is_set():
                    # Delete any stacks that have already completed
                    self._delete_stacks(
                        filter(lambda x: x.created, self.providers)
                    )

        if error.is_set():
            # Stacks are being deleted because one errored.  Sit and wait until
            # they are all completed.
            self._monitor_stack_deletion()



import random
from http import HTTPStatus
from io import IOBase, BytesIO
from time import sleep
from bson_extra import bson_extra

from pyqalx.core.encryption import QalxEncrypt

try:
    from json.decoder import JSONDecodeError
except ImportError:
    # JSONDecodeError is for Python3 only
    JSONDecodeError = ValueError

import requests


class PyQalxAPIException(Exception):
    """
    A generic error for PyQalxAPI
    """


class BasePyQalxAPI(object):
    """
    Core functionality for the transport layer.  Handles the building of the
    request and the response
    """

    def __init__(self, session):
        self.session = session
        self.config = self.session.config
        self.base_url = self.config.get("BASE_URL", "https://api.qalx.net/")
        self.__token = None
        self.token = self.config["TOKEN"]
        self._last_response = None
        self._host_ip = None

        # The maximum amount of retries for specific response status codes
        self._max_retries = {
            HTTPStatus.TOO_MANY_REQUESTS.value: {"retries": 0},
            HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                "retries": 0,
                "max": self.config["MAX_RETRY_500"],
            },
        }

    def _should_retry(self, status_code):
        """
        Determines whether this request should be retried based on the
        status code and the number of times this request has already been tried
        :param status_code:The status code of the response
        :type status_code:int
        :return:bool of whether or not the request should be retried.
        """
        _too_many_requests = HTTPStatus.TOO_MANY_REQUESTS.value
        _internal_server_error = HTTPStatus.INTERNAL_SERVER_ERROR.value
        should_retry = False

        if status_code == _too_many_requests:
            # Throttled.  Always retry using a random expanding time window,
            # log a warning
            should_retry = True
            too_many_requests = self._max_retries[_too_many_requests]

            too_many_requests["retries"] += 1
            # Sleep for a random increasing window to allow the throttle to
            # expire
            sleep_for = too_many_requests["retries"] * random.randint(1, 5)

            self.session.log.warning(
                f"Got `{status_code}` response from API. Sleeping for "
                f"`{sleep_for}` seconds before trying again."
            )
            sleep(sleep_for)
        elif status_code >= _internal_server_error:
            # Error from API or requests exception.  Only try
            # up to the amount specified in the config
            internal_server_error = self._max_retries[_internal_server_error]
            if internal_server_error["retries"] < internal_server_error["max"]:
                should_retry = True
                internal_server_error["retries"] += 1
        return should_retry

    @property
    def token(self):
        """
        Returns the current API Token that is in use for this session

        :return: str
        """
        return self.__token

    @token.setter
    def token(self, token):
        """
        Setter for the token attribute.  Whenever a token is set the
        requests.Session object should be reinitialised to ensure that it
        has the correct token cached

        :param token: The token to use for authing all requests
        """
        self.__token = token
        self._initialise_session()

    def _initialise_session(self):
        """
        Initialises a new instance of `requests.Session` and sets the
        authentication headers.  This needs to happen post fork for Workers,
        due to multiprocessing issues - https://github.com/psf/requests/issues/4323
        """
        self._session = requests.Session()
        self._session.headers = {
            "Authorization": f"Token {self.token}",
            "Content-Type": "application/json",
        }

    def _build_request(self, url, method, include_auth_headers=True, **kwargs):
        """
        Handles making the request
        :param url: The url to make the request to
        :param include_auth_headers: Should we include the auth headers
        :param kwargs: kwargs to pass through to the request
        :return: response
        """
        headers = kwargs.pop("headers", {})
        if include_auth_headers is False:
            # We might not include the Authorization header if we are PUTing to S3
            headers.update({"Authorization": None})

        for k in self._max_retries.keys():
            # Reset all retries.
            self._max_retries[k]["retries"] = 0

        should_retry = True

        while should_retry:
            # There may be instances where the API returns a 500 error.  This
            # could be due to many things (network issues, cold starts etc).
            # To avoid terminating workers we retry up to `MAX_RETRIES` times
            # to make our best attempt at completing the request successfully.
            try:
                resp = self._session.request(
                    url=url, method=method, headers=headers, **kwargs
                )
                self._last_response = resp
                status_code = resp.status_code
            except requests.exceptions.RequestException as exc:
                # RequestException: The base exception that `requests` can
                #                   raise. Catch everything and return to the
                #                   client
                resp = exc
                status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
            # Check whether this request should be retried
            should_retry = self._should_retry(status_code)
        return self._build_response(resp)

    def _build_non_safe_request(self, method, endpoint, **kwargs):
        """
        Helper method for performing non safe requests (POST, PUT, PATCH etc).
        Handles the possibility of a user trying to create or update an item
        that also has a file
        :param method: The HTTP method
        :param endpoint: The endpoint you want to query: 'item`
        :param kwargs: Any data you want to pass to the request.
                      Use `json` key for data you want to post. Anything else
                      will be passed to `requests` as kwargs
                      json={'data': {'some': 'data'}, 'meta': {'some': 'meta'}}
        :return: response
        """
        url = self._get_url(endpoint)
        data = kwargs.get("data", None)

        if data:
            # data is dumped to json in a way that enables type to be preserved.
            # This means that we cannot use the `json` keyword argument so
            # have to pass the data using `data` and set the Content-Type of
            # the request to `application/json`
            data = bson_extra.dumps(data)
            if not isinstance(data, bytes):
                data = data.encode("utf-8")
            kwargs["data"] = data
        resp_ok, data = self._build_request(url=url, method=method, **kwargs)
        return resp_ok, data

    def _build_external_request(self, url, method, **kwargs):
        """
        Helper for building a request that is to an external source
        :param url: The URL to request
        :param method: The method to call
        :return: The response from `self._build_request`
        """
        return self._build_request(
            url=url, method=method, include_auth_headers=False, **kwargs
        )

    @staticmethod
    def _build_response(resp):
        """
        Builds the response that gets sent back to the client
        :param resp: The response from `requests` or an instance
                     of `RequestException`
        :return: tuple of `(response_ok:bool, response data:dict)`
        """
        # If it doesn't have `ok` then a `requests` exception was raised
        is_ok = getattr(resp, "ok", False)
        try:
            # Ensure that we load data to the correct type
            data = resp.json(**bson_extra._bson_loads_kwargs())
        except (AttributeError, JSONDecodeError):
            # AttributeError: `resp` is an instance of RequestException
            # JSONDecodeError: `resp` is a normal response but has no data.
            #                   i.e. a 500 or a `delete` response
            data = []
        if is_ok is False:
            # If either of these are missing then a `requests` exception was
            # raised
            status_code = getattr(resp, "status_code", "")
            reason = getattr(resp, "reason", resp)
            data = {
                "status_code": status_code,
                "reason": reason,
                "errors": data,
            }
        return is_ok, data

    def _get_url(self, endpoint):
        """
        Builds the URL for the request from the base_url and the endpoint
        :param endpoint: The endpoint you want to query: 'item/<item_guid>'
        :return: url
        """
        url = "{base_url}{endpoint}".format(
            base_url=self.base_url, endpoint=endpoint.rstrip("/").lstrip("/")
        )
        return url

    def _upload_to_s3(self, data, source, file_key):
        """
        Given a response object with a file url in will attempt to
        PUT the file onto S3

        :param data:        The data from the api
        :param source:  The path to the file to upload or the file itself
        :param file_key:    The file key in the data dictionary
        :return: response
        """
        url = data[file_key]["put_url"]

        if self.is_filestream(source):
            # seek(0) because the file could be at the end if it has already
            # been uploaded (i.e. if add_many uses the same file for
            # multiple entities)
            source.seek(0)
            file_data = source.read()
        else:
            file_data = open(source, "rb").read()
        # Check if we need to encrypt the file data
        if data[file_key].get("keyfile") is not None:
            key_file = self.config["KEYFILE"]
            qalx_encrypt = QalxEncrypt(key_file)
            token = qalx_encrypt.encrypt(bytes(file_data))
            file_data = BytesIO(token)
        return self._build_external_request(
            url=url, method="PUT", data=file_data
        )

    def _get_external_ip(self):
        """
        Gets the IP external address
        """
        if self._host_ip is None:
            # Get the external IP and cache it on the session
            success, ip = self._build_external_request(
                url="https://api.myip.com/", method="GET"
            )
            if success:
                self._host_ip = ip["ip"]
        return self._host_ip

    @staticmethod
    def is_filestream(input_object):
        """
        Determines whether the given input object is a file stream or a filepath

        :param input_object: An instance of IOBase or a filepath
        :type: io.IOBase or str
        :return: bool
        """
        return isinstance(input_object, IOBase)



"""
Factory adapter module. Contains the main class that provides factory related
interaction with the qalx API: QalxFactory
"""

import json

from ...factories.mixins import FactoryValidateMixin
from ..entities.factory import Factory
from .adapter import QalxFileAdapter
from ..errors import QalxFactoryValidationError, QalxAPIResponseError


class QalxFactory(QalxFileAdapter):
    """
    Provides an interface for accessing the API for
    :class:`~pyqalx.core.entities.factory.Factory` entities
    """

    _entity_class = Factory
    _file_key = "plan"
    # The `status` key should be saved for this adapter
    _keys_to_exclude = {"guid", "info"}

    def validate(self, plan_path):
        """
        Reads a .yaml file that contains the definition of the factory plan and
        then runs the factory validation endpoint in the API

        :param plan_path:   A local path to the .yaml file that contains the
                            factory plan definition
        :type plan_path: str

        :raises QalxFactoryValidationError:     If the .yaml file does not
                                                exist, if there are any errors
                                                parsing the yaml file to a
                                                dict, or if the yaml plan
                                                structure is invalid
        """
        plan = FactoryValidateMixin.read_plan(plan_path)
        plan_json = json.dumps(FactoryValidateMixin.load_plan(plan))
        try:
            self._process_api_request(
                "post", "factory/validate", plan=plan_json
            )
        except QalxAPIResponseError as exc:
            raise QalxFactoryValidationError(
                f"There was an error validating the factory plan:\n\n{exc}"
            )

    def add(
        self, name, source=None, file_name="", meta=None, upload=True, **kwargs
    ):
        """
        Adds a `Factory` instance

        :param name:        The name of the factory
        :type  name:        str
        :param source:  file path to the factory plan or instance of
                            StringIO or BytesIO
                            (https://docs.python.org/3/library/io.html)
        :type  source:  str or `io` object
        :param file_name:   Name for the factory plan input file. Optional if a
                            file path is given for source parameter.
                            Required if an `io` object
        :type  file_name:   str
        :param meta:        A dictionary of metadata to store
        :type  meta:        dict
        :param upload: Whether the file should be automatically uploaded or not
        :type upload: bool

        :return: A newly created `Factory` instance
        """

        return super(QalxFactory, self).add(
            name=name,
            source=source,
            file_name=file_name,
            meta=meta,
            file_key=self.entity_class._file_key,
            as_file=False,
            upload=upload,
            encrypt=False,
            **kwargs,
        )

    def save(
        self, entity, source=None, file_name="", upload=True, *args, **kwargs
    ):
        """
        Saves an updated existing `Factory` instance.

        :param entity: The entity that we are saving
        :type  entity: An instance of `Factory`
        :param source:  file path to the factory plan or instance of
                            StringIO or BytesIO
                            (https://docs.python.org/3/library/io.html)
        :type  source:  str or `io` object
        :param file_name:   Name for the factory plan input file. Optional if a
                            file path is given for source parameter.
                            Required if an `io` object
        :type  file_name:   str
        :param upload: Whether the file should be automatically uploaded or not
        :type upload: bool
        :return: An updated `Factory` instance
        """
        return super(QalxFactory, self).save(
            entity=entity,
            source=source,
            file_name=file_name,
            as_file=False,
            upload=upload,
            encrypt=False,
            *args,
            **kwargs,
        )

    def get(self, guid, *args, **kwargs):
        # extended to create better autodocs
        return super(QalxFactory, self).get(guid, *args, **kwargs)

    def find(
        self,
        query=None,
        sort=None,
        skip=0,
        limit=25,
        many=True,
        include_session_tags=True,
        **kwargs,
    ):
        # extended to create better autodocs
        return super(QalxFactory, self).find(
            query=query,
            sort=sort,
            skip=skip,
            limit=limit,
            many=many,
            include_session_tags=include_session_tags,
            **kwargs,
        )

    def find_one(self, query, **kwargs):
        # extended to create better autodocs
        return super(QalxFactory, self).find_one(query=query, **kwargs)

    def reload(self, entity, *args, **kwargs):
        # extended to create better autodocs
        return super(QalxFactory, self).reload(entity=entity, **kwargs)

    def aggregate(self, aggregate, **kwargs):
        # extended to create better autodocs
        return super(QalxFactory, self).aggregate(aggregate=aggregate, **kwargs)



"""
Mix-in classes for the main factory class. Each one of the classes in this module
contains functionality for a specific factories related stage. Keeping methods
contained in these mix-ins keeps the main factory class cleaner
"""

import concurrent.futures
import logging
import shutil
from subprocess import Popen, PIPE, CalledProcessError
import sys
import uuid
from time import sleep

from pyqalx.factories.debug import FactoryPlanDebug

try:
    import git
except ImportError:
    # ImportError: Most likely git is not in the expected place.  Not an issue
    #   unless the user wants to use bot_sources.GitBotSource
    pass
import os
import stat
import yaml
from yaml import YAMLError

from bson_extra import bson_extra

from pyqalx.config import UserConfig, BotConfig
from pyqalx.factories.bot_sources import BotSource, QalxItemBotSource
from pyqalx.core.entities.factory import Factory as FactoryEntity
from pyqalx.core.errors import (
    QalxFactoryPackError,
    QalxFactoryBuildError,
    QalxFactoryError,
    QalxAPIResponseError,
)
from pyqalx.core.errors import QalxFactoryValidationError
from pyqalx.factories.sectors.manager import SectorManager
from pyqalx.factories.sectors.sectors import Sectors
from pyqalx.factories.sectors.providers import (
    _providers,
    LocalProvider,
    AWSProvider,
)
from pyqalx.factories.tree import QalxTree
from pyqalx.core.utils import _msg_user

default_aws_profile = AWSProvider.default_profile_name()


class FactoryValidateMixin:
    """Specific Validate functionality"""

    @staticmethod
    def load_plan(plan_yml):
        """
        Loads the plan from yaml to JSON
        :param plan_yml: The plan as a yaml string
        :return: The plan as JSON
        :raises QalxFactoryError if there was a problem parsing the yaml to JSON
        """
        try:
            return yaml.safe_load(plan_yml)
        except YAMLError as exc:
            raise QalxFactoryError(
                f"There was an error parsing the plan yaml file:\n\n{exc}"
            )

    @staticmethod
    def read_plan(plan_path):
        """
        Reads a .yaml file that contains the definition of the factory plan and
        then returns it as a yaml string

        :param plan_path:   A local path to the .yaml file that contains the
                            factory plan definition

        :return:    The plan as a yaml string

        :raises QalxFactoryValidationError:     If the .yaml file does not
                                                exist
        """
        if not os.path.exists(plan_path):
            raise QalxFactoryValidationError(
                f"The plan path: {plan_path} was not found"
            )
        with open(plan_path, "r") as plan_file:
            plan = plan_file.read()
        return plan


class FactoryPackMixin:
    """Specific Pack functionality"""

    def _create_build_path(self):
        """
        Creates the build path where the bot code will live.  By default
        this will use a random uuid as we won't have a Factory entity as the
        point of packing.

        :raises QalxFactoryPackError: if
                 the directory can not be created.
        """
        try:
            os.makedirs(self.build_path)
        except (OSError, PermissionError) as exc:
            raise QalxFactoryPackError(
                f"There was an error creating the build directory\n\n{exc}"
            )

    def _bots_for_pack(self, stage=None):
        """
        Returns the bots key from the plan.  If `stage` is provided then return
        only those bots that match that stage

        :param stage: If provided, will return the bots in `factory:bots` on
        the plan that match the given stage.  If not provided, returns all bots
        :type stage: str
        :return: A dict of `bot_name:bot_source`
        """
        bots = self._plan_bots
        if stage is not None:
            # A stage was provided, only return bots that match that specific
            # stage
            _stage = self._plan_stages[stage]
            bot_names = set()
            for sector in _stage.values():
                sector_bots = isinstance(sector, dict) and sector.get("bots")
                if sector_bots:
                    bot_names.update(sector_bots.keys())
            bots = dict(filter(lambda i: i[0] in bot_names, bots.items()))
        return bots

    def _download_bot(self, bot_name, bot):
        """
        Attempts to download the given bot from the given source.

        :param bot_name:The name of the bot that is being downloaded
        :param bot:The dict representing the source bot from the plan
        """

        try:
            BotSourceClass = BotSource.get_source(bot)
        except KeyError as exc:
            raise QalxFactoryPackError(exc)
        bot_source = BotSourceClass(
            bot_name, bot, self.build_path, log=self.session.log
        )
        bot_source.download_bot()
        self.bot_sources.append(bot_source)

    def _download_bots(self, bots):
        """
        Iterates through the given bots and attempts to download them.
        Catches any potential error that could be raised and reraises as a
        QalxFactoryPackError for consistency

        :param bots:    A dictionary of {bot_name:bot_source}
        :type  bots:    dict
        :raises QalxFactoryPackError:
        """
        results = []
        exception = None
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            for bot_name, bot in bots.items():
                results.append(
                    executor.submit(self._download_bot, bot_name, bot)
                )
            for result in concurrent.futures.as_completed(results):
                try:
                    # Check if there was an error downloading the bot
                    result.result()
                except (
                    git.GitCommandError,
                    CalledProcessError,
                    NotADirectoryError,
                    PermissionError,
                    FileNotFoundError,
                    QalxFactoryPackError,
                    QalxFactoryValidationError,
                ) as exc:
                    # Provides a common way of handling all the potential errors
                    # that could occur when packing bots

                    # git.GitCommandError: There was an error clone the git repo
                    # subprocess.CalledProcessError: There was an error pip
                    #                                downloading the package
                    # NotADirectoryError: The specified file wasn't a directory
                    # FileNotFoundError: The local path could not be found
                    # PermissionError: No permission to read from directory
                    # QalxFactoryValidationError: Error validating bot path

                    # Grab this exception and raise later.  Need to wait for
                    # all processing to finish before exiting otherwise it won't
                    # be possible to tidy up properly as a thread could be
                    # writing to the directory as the deletion occurs
                    exception = exc

        if exception is not None:
            # Ensure that the packing tidies up anything that was successful
            # before this failure
            self._delete_build_path()
            raise QalxFactoryPackError(exception)


class FactoryBuildMixin:
    """
    Specific build functionality
    """

    @staticmethod
    def _metadata(entity, plan_bot_name=None):
        """
        The base metadata that should be stored on any entities that are linked
        to this factory
        """
        metadata = {
            "_class": "pyqalx.factory",
            # Namedspaced metadata to avoid potential name clashing
            # with any custom metadata provided by the user.
            BaseFactory.PYQALX_METADATA_NAMESPACE_KEY: {
                BaseFactory.PYQALX_METADATA_STAGE_KEY: entity.stage,
                BaseFactory.PYQALX_METADATA_FACTORY_GUID_KEY: entity.guid,
            },
        }
        if plan_bot_name is not None:
            # This enables identification between Bot entities
            # and Item entities that contain the Bot code.
            metadata[BaseFactory.PYQALX_METADATA_NAMESPACE_KEY][
                BaseFactory.PYQALX_METADATA_PLAN_BOT_NAME_KEY
            ] = plan_bot_name
        return metadata

    def _add_entity(self, plan):
        """
        Helper for adding a factory entity to the API.
        :param plan: The path to the plan file
        :type plan:str
        :return: ~pyqalx.core.entities.factory.Factory
        """
        factory_meta = self._plan_meta
        factory_meta[self.PYQALX_METADATA_NAMESPACE_KEY] = {
            self.PYQALX_METADATA_BUILD_PATH_KEY: os.path.basename(
                self.build_path
            )
        }
        self.entity = self.session.factory.add(
            name=self._plan_name,
            source=plan,
            stage=self.stage,
            meta=factory_meta,
        )
        return self.entity

    def _create_bot_items(self):
        """
        Creates a qalx item with the code and metadata about the code of each
        bot on a factory stage. For each bot and bot source, a qalx item is
        created with the following specification:

            source: An input file is provided and it is the zipped
                        directory of the bot source code
            meta:       The metadata, with the following key/value pairs:
                "_class":           "pyqalx.factory"
                "stage:             The stage that the bots belong to
                "deployment_id":    The factory deployment id
                "md5_hash":         The md5 checksum of the contents of the bot
                                    source code directory
        """
        item_dict_list = []
        hashes = {}
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            for bot_source in self.bot_sources:
                hashes[executor.submit(bot_source.get_hash)] = bot_source
            for future in concurrent.futures.as_completed(hashes):
                bot_source = hashes[future]
                plan_bot_name = bot_source.bot_name
                meta = self._metadata(
                    entity=self.entity, plan_bot_name=plan_bot_name
                )
                meta["md5_hash"] = bot_source.md5_hash
                item = {"meta": meta}
                zip_path = bot_source.zip_build_path()
                if zip_path:
                    # pypi bots won't have a file
                    item["source"] = zip_path
                item_dict_list.append(item)
        self.bot_items = self.session.item.add_many(item_dict_list).items_

    @staticmethod
    def _sanitise_cli_key(k):
        """
        For the given arguments given to `bot-start` we need to ensure that
        the keys are sanitised in case the user provides a key with either `--`
        or `-` in the plan (or they didn't provide either)
        :param k: The key to sanitise
        :type k: str
        :return: The sanitised key - with the correct number of hyphens
        prefixing it depending on whether they gave the short or longform
        argument
        """
        if k.startswith("--"):
            k = k[2:]
        elif k.startswith("-"):
            k = k[1:]
        if len(k) > 1:
            return f"--{k}"
        return f"-{k}"

    @staticmethod
    def _environment_variable_key(parameter_name):
        """
        From the given parameter name, will return the name that the environment
        variable should be
        :param name:The given parameter name in the format:
        "<factory_guid>-<ec2instancename>_<CONFIG_KEY_PREFIX>_<PARAMETER_NAME>"
        :return:"<CONFIG_KEY_PREFIX>_<PARAMETER_NAME>
        """
        return parameter_name.split("_", 1)[-1]

    def bootstrap_environment(self, ProviderClass, region_name, instance_name):
        """
        Bootstraps the environment by getting all the parameters stored on the
        cloud provider and saving them to the environment for the config to find
        :param ProviderClass:The class of the provider that is bootstrapping
        the bots
        :type ProviderClass:~factories.factory.sectors.providers.provider.BaseProvider
        :param region_name:The remote region that is used for bootstrapping the instance
        :type region_name:str
        :param instance_name:The name of the instance used when creating the stack
        :type instance_name:str
        """
        provider = ProviderClass(
            factory=self,
            # We won't have the stack or profile on remote sectors at this
            # point as we won't have been able to query the API yet to get the data.
            stack={},
            profile_name=None,
            region_name=region_name,
        )
        # There will be no config at this point so just use the
        # raw factory_guid
        provider.set_environment_variables(instance_name=instance_name)
        # Reload the current session so that it has the updated values
        # from the environment variables.  skip_ini is True because config
        # will be loading from the environment variables.  Ensure config_class
        # remains the same
        self.session = self.session.__class__(
            config_class=self.session.config.__class__, skip_ini=True
        )

    def bootstrap_bots(
        self,
        factory_guid,
        sector_name,
        profile_name,
        region_name,
        instance_name,
        ProviderClass,
    ):
        """
        Iterates through all the bots on the provided sector_name and downloads
        the code from the Items on qalx.  Once this is successful will then
        start the bots in the same way as bots on local sectors are started.
        :param factory_guid:The factory_guid that should be bootstrapped
        :param sector_name: The sector name that contains the bots that are
        being bootstrapped
        :type sector_name: str
        :param profile_name: The profile_name of the sector being bootstrapped
        :param region_name: The region_name of the sector being bootstrapped
        :param instance_name: The name of the instance being bootstrapped.
        The name must match the name used for parameters that are created
        during stack creation
        :param ProviderClass:The class of the provider that is doing the
        bootstrapping
        :type ProviderClass: A subclass of ~providers.provider.BaseProvider
        """
        self.bootstrap_environment(
            ProviderClass=ProviderClass,
            region_name=region_name,
            instance_name=instance_name,
        )
        # Get the factory and load the sectors
        self.entity = factory_guid
        sectors = Sectors(
            self, validate_default_region=False, validate_permissions=False
        )

        sector = sectors[ProviderClass._type][profile_name][region_name][
            sector_name
        ]
        # All the bot names across the approriate sectors.  A single sector
        # could have multiple bots
        bot_names = list(sector["bots"].keys())

        # Ensure that the build_path gets created so that the bot code can be
        # downloaded.  Ensure the build path is deleted first.  It could exist
        # if the instance has been rebooted - in which case we act as if it's
        # the first boot to handle updates more easily.
        self._delete_build_path()
        self._create_build_path()

        namespace_key = self.PYQALX_METADATA_NAMESPACE_KEY
        factory_guid_key = self.PYQALX_METADATA_FACTORY_GUID_KEY
        plan_bot_name_key = self.PYQALX_METADATA_PLAN_BOT_NAME_KEY
        # Get just the code items that relate to the bots we are bootstrapping
        query = f"metadata.data.{namespace_key}"
        items = self.session.item.find(
            query={
                f"{query}.{factory_guid_key}": self.entity.guid,
                f"{query}.{plan_bot_name_key}": {"$in": bot_names},
            }
        )

        # Iterate through all the bot code items that we are starting the bots
        # for, download the code and then start the bot (async)
        results = []
        start_context = {}
        max_workers = items.query.count
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=max_workers
        ) as executor:
            for bot_code_item in items.data:
                bot_name = bot_code_item.meta[namespace_key][plan_bot_name_key]
                bot = self._plan_bots[bot_name]
                bot_source = QalxItemBotSource(
                    bot_name=bot_name,
                    bot=bot,
                    build_path_dir=self.build_path,
                    bot_code_item=bot_code_item,
                    log=self.session.log,
                )
                results.append(executor.submit(bot_source.download_bot))

                # Context required when the download is completed so that the
                # bot can be started
                bot_arguments = sector["bots"][bot_name]

                start_context[bot_name] = {
                    "build_path": bot_source.build_path,
                    "plan_bot_name": bot_name,
                    "bot_path": bot["bot_path"],
                    "bot_arguments": bot_arguments,
                    "sector": sector,
                    "skip_ini": True,
                }

            for future in concurrent.futures.as_completed(results):
                # As the downloads complete, start the bots up.
                output, error = self.start_bot(**start_context[future.result()])
                if output.returncode is not None:
                    # There was an error starting up this bot. Raise an
                    # exception as this needs to be handled by the bootstrap
                    # script itself which will take appropriate action.
                    raise QalxFactoryBuildError(str(error))

    def start_bot(
        self,
        plan_bot_name,
        build_path,
        bot_path,
        bot_arguments,
        sector,
        skip_ini=False,
    ):
        """
        Starts a single bot in an orphaned subprocess. Enables factories to start
        multiple bots without blocking.
        :param plan_bot_name:The name of the bot from the plan
        :type plan_bot_name:str
        :param build_path:The path where the bot source resides
        :type build_path: str
        :param bot_path:The path to the bot to be used in the start call
        :type bot_path:str
        :param bot_arguments:Any additional parameters to be passed to start
        :type bot_arguments: dict
        :param sector:The specific sector that contains this bot
        :type sector: dict
        :param skip_ini:Should the initialisation of the qalx session be skipped?
        :type skip_ini:bool
        :return:The output of the subprocess call and any errors that occurred.
        """
        arguments = []

        for k, v in bot_arguments.items():
            arguments += [FactoryBuildMixin._sanitise_cli_key(k), str(v)]

        if skip_ini:
            # Remote bots won't have access to a config file.
            arguments.append("--skip-ini")

        # All bots should be started but in a "stopped" state.
        # This will enable all bots to be created before any start
        # processing data
        arguments.append("--stop")
        metadata = FactoryBuildMixin._metadata(
            entity=self.entity, plan_bot_name=plan_bot_name
        )
        stage_workflow = self._plan_stage.get("workflow")
        if stage_workflow:
            plan_workflow = self._plan_workflows[stage_workflow]
            successors = plan_workflow.get_successors(plan_bot_name)

            # If the bot isn't part of the workflow, don't add the extra args
            if successors:
                for successor in successors:
                    arguments += ["--workflow", successor]

        for key, value in metadata.items():
            # Builds the meta data to be saved on the bot
            if not isinstance(value, str):
                # This value isn't a string.  Dump to JSON so it can be saved it
                # in the correct structure
                value = bson_extra.dumps(value)
            arguments += ["--meta", f"{key}={value}"]

        # subprocess will be called using the same environment as the main
        # process.  However, this won't know about the path for the bot as
        # this gets updated at runtime .  Add the package on the build path
        # (just for the context of this subprocess
        # call).  We then need to pass the full environment through to
        # subprocess along with the modified path so that the bot can be
        # imported.
        modified_env = os.environ.copy()
        modified_path = sys.path + [build_path]
        python_path = os.pathsep.join(modified_path)
        modified_env["PYTHONPATH"] = python_path

        qalx = shutil.which("qalx")

        output = Popen(
            [
                qalx,
                "--user-profile",
                sector[Sectors.USER_PROFILE_KEY],
                "--bot-profile",
                sector[Sectors.BOT_PROFILE_KEY],
                "bot-start",
                *arguments,
                bot_path,
            ],
            # Redirect stderr and stdout to PIPE so we can capture
            # errors to give to the user in the main thread
            stdout=PIPE,
            stderr=PIPE,
            env=modified_env,
        )

        # Arbitrary timeout to give the bot a chance to start.
        # TODO: Handle getting the bot and checking the status of the
        #  workers to ensure it's been created successfully.  Will allow quicker
        # startups
        sleep(10)
        # .poll() will populate the returncode for us - but nothing else
        output.poll()
        error = None
        if output.returncode is not None:
            # Some sort of error occurred when starting this bot
            stdout, stderr = output.communicate()
            # There might be an error on `stdout` if there was an exception
            # raised i.e. in the config if the log file path
            # couldn't be created.  Therefore, join stdout and stderr together
            # (giving more prominence to stderr) just in case there are issues
            # in two places.
            error = "\n".join([stderr.decode("utf-8"), stdout.decode("utf-8")])
        return output, error

    def _create_stacks(self):
        """
        Creates the stacks for each bot. A stack is all the resources for the
        bots on a sector with an intersection of profile_name/region_name.  All
        local bots will be on their own stack, whereas remote bots will be
        separated out into separate sectors on stacks which share a
        profile_name/region_name.
        When this completes successfully all the bots will have been started in
        a stopped state.
        If there are any issues with creating a stack (i.e. stack creation fails,
        bot fails to start etc) then all stacks get deleted and the user
        is given an error.
        """
        providers = {}
        stack_mapping = {}
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            # Synchronously start the processes.  Local bots will have a sleep
            # in the main thread but that shouldn't cause remote stacks to
            # not be started as soon as possible.
            for _provider in self.stacks:
                if _provider is not None:
                    # Provider will be None if a provider is not configured
                    # properly
                    providers[
                        executor.submit(_provider.create_stack)
                    ] = _provider
            for future in concurrent.futures.as_completed(providers):
                provider = providers[future]
                for sector_name in provider.stack.keys():
                    # Ensure the sectors are saved against the stack ids
                    # - only remote sectors are saved.
                    if not isinstance(provider, LocalProvider):
                        stack_mapping[sector_name] = provider.stack_id

        # Add the stack mappings to the entity for later saving.
        self.entity["stacks"] = stack_mapping
        all_completed = False
        if providers:
            # Only run the manager if there are providers - if there aren't
            # any then there was a config error with one/all the providers
            manager = SectorManager(
                providers=list(providers.values()), plan_debug=self._plan_debug
            )
            manager._monitor_stack_creation()
            all_completed = manager._all_completed()

        if all_completed or self._plan_debug._demolish_on_failure is False:
            # Add the stack mappings to the entity if the stacks
            # created successfully or if we the debug key is set to not demolish
            # on failure.
            self.entity["stacks"] = stack_mapping
            self.session.factory.save(self.entity)
        return all_completed

    def _add_factory_bots(self):
        """
        Updates the factory entity with the Bot entities and the
        Item entities that the Bot code is stored on.  The `plan_bot_name` is
        stored in metadata on the Item and the Bot which can be used to figure
        out how they are related.
        """
        namespace_key = self.PYQALX_METADATA_NAMESPACE_KEY
        plan_bot_name_key = self.PYQALX_METADATA_PLAN_BOT_NAME_KEY
        item_bot_name_to_guid = {
            item.meta[namespace_key][plan_bot_name_key]: item.guid
            for item in self.bot_items
        }

        expected_build_bots = []
        for bot_guid, plan_bot_name in self.bots.keys():
            # There could be more bot entities than item entities if the same
            # bot is used on multiple sectors.
            expected_build_bots.append(
                {"code": item_bot_name_to_guid[plan_bot_name], "bot": bot_guid}
            )
        self.entity["bots"] = expected_build_bots
        self.session.factory.save(self.entity)

    def _resume_bots(self):
        """
        Once all the bots are successfully created then they should be resumed.
        """
        for bot in self.bots.values():
            self.session.bot.resume(bot)


class FactoryDemolishMixin:
    """
    Specific Demolish functionality

    Class Attributes
        MAX_BOT_TERMINATE_PROCESSES:    Maximum number of processes to allow for
                                        terminating bots
    """

    MAX_BOT_TERMINATE_PROCESSES = 4

    def get_workflow(self):
        """
        Get the workflow according to which the factory has been deployed. If no
        workflow has been specified for the deployed stage of the factory, then
        a "flat" workflow with all the bots on the factory entity at the same
        level is returned

        :return: The workflow of the deployed stage which is a QalxTree instance
        """
        workflow_key = self._plan_stage.get("workflow")
        bot_names = list(
            set([bot_plan_name for _, bot_plan_name in self.bots.keys()])
        )
        if workflow_key:
            workflow = self._plan_workflows[workflow_key]
            # We need to add leaf nodes to the tree for all the bots that are on
            # the stage, but not on the workflow definition from the plan. These
            # bots are at the top level of the tree
            for bot_name in bot_names:
                if bot_name not in workflow.nodes:
                    workflow.create_node(
                        bot_name, bot_name, parent=workflow.root
                    )
            return workflow
        return QalxTree({"flat_workflow": bot_names})

    def wait_on_bot_termination(self, bot):
        """
        For a provided bot entity, iterates through all the workers and returns
        only after it has been asserted that all workers have been warm
        terminated.

        :param bot:     The bot to wait on until all workers are warm terminated
        """

        def _terminate_warm(_bot):
            # Helper for returning the `terminate_warm` signal of all works
            # on a given bot
            return [
                self.session.worker.signal_class(_worker).terminate_warm
                for _worker in _bot.workers
            ]

        terminated_signals = _terminate_warm(bot)

        while not all(terminated_signals):
            # sleep, to prevent hitting the API too many times while waiting
            # for termination.
            sleep(5)
            bot = self.session.bot.reload(bot)
            terminated_signals = _terminate_warm(bot)

    def safe_terminate_bot(self, plan_bot_name):
        """
        Safely terminate one or more bots with the provided plan name. Calls
        bot.terminate with warm=True on the specified bots and then for each of
        the bots calls wait_on_bot_termination.  There may be multiple bots
        with the same name - i.e. same bot on different sectors.  These can
        be safely terminated in parallel as this method is only called with
        the correct bot names for the correct point in the workflow
        :param  plan_bot_name:   The name of the bot to terminate on the plan
        :type   plan_bot_name    str
        :return bot_name
        """
        self._msg_user(msg=f"[{plan_bot_name}] terminating")
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.MAX_BOT_TERMINATE_PROCESSES
        ) as executor:
            for (_, _plan_bot_name), bot in self.bots.items():
                if _plan_bot_name == plan_bot_name:
                    # Send termination signal to all affected bots.
                    # and wait for termination to complete.
                    executor.submit(
                        self.session.bot.terminate, entity=bot, warm=True
                    )
                    executor.submit(self.wait_on_bot_termination, bot)
        self._msg_user(msg=f"[{plan_bot_name}] terminated successfully")
        return plan_bot_name

    def concurrent_terminate(self, workflow, bot_names):
        """
        Terminate a list of bots concurrently with concurrent.futures

        :param  workflow:   The workflow that has to be respected in termination
        :type   workflow:   QalxTree instance
        :param  bot_names:  The bot names to terminate
        :type   bot_names:  list
        """
        results = []
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.MAX_BOT_TERMINATE_PROCESSES
        ) as executor:
            for bot_name in bot_names:
                results.append(
                    executor.submit(self.safe_terminate_bot, bot_name)
                )
            for finished in concurrent.futures.as_completed(results):
                bot_name = finished.result()
                try:
                    self.concurrent_terminate(
                        workflow, workflow.get_successors(bot_name)
                    )
                except QalxAPIResponseError as exc:
                    # QalxAPIResponseError:  Issue when terminating bot.
                    #                        Unlikely to happen if the user gets
                    #                        this far but they should be made
                    #                        aware so they can take manual steps
                    self._msg_user(
                        msg=f"Exception occurred when "
                        f"terminating {bot_name}.  "
                        f"You may need to terminate this "
                        f"manually: {exc} ",
                        level=logging.ERROR,
                    )

    def breadth_first_terminate(self):
        """
        Breadth-first terminate of bots for a provided tree structure of a
        workflow. Traverses the nodes of the workflow in breadth-first order and
        calls "concurrent_terminate" on the successors of each node.

        Once all the bots are terminated the stacks can be deleted.  For remote
        stacks the bot processes will exit when the stack is deleted.  For local
        stacks, the bots will just exit normally as if `qalx bot-terminate` had
        been called on them

        """
        workflow = self.get_workflow()
        successors = workflow.get_successors(workflow.root)
        self.concurrent_terminate(workflow, successors)


class BaseFactory(
    FactoryValidateMixin,
    FactoryPackMixin,
    FactoryBuildMixin,
    FactoryDemolishMixin,
):
    """
    Base factory class. Contains attributes and methods that are not tied to a
    specific stage of a factory

    Class attributes
        PYQALX_METADATA_NAMESPACE_KEY:          Metadata key for pyqalx namespace
        PYQALX_METADATA_STAGE_KEY:              Metadata key for stage
        PYQALX_METADATA_FACTORY_GUID_KEY:       Metadata key for factory guid
        PYQALX_METADATA_PLAN_BOT_NAME_KEY:      Metadata key for plan bot name

    Instance attributes
        session:        A QalxSession instance
        user_profile:   The qalx profile to use as a user profile.
        bot_profile:    The qalx profile to use for starting bots. Default value
                        is according to the default bot configuration
        aws_profile:    The AWS profile to use for remote sectors. Default value
                        is according to boto3 defaults
        __entity: The Factory entity from the API.
        __plan:           A dictionary with the details of the plan of the factory
        __stage: The stage that the build/pack is using.  Uses the stage on the

        build_path:         The root path that the bot code gets downloaded to
                            and built from
        bot_sources:    A list of BotSource instances for each of the bots
                            that are being build
        _bots: A mapping of the bot names on the plan
              to the ~pyqalx.core.entities.bot.Bot instances
    """

    PYQALX_METADATA_NAMESPACE_KEY = "pyqalx"
    PYQALX_METADATA_STAGE_KEY = "stage"
    PYQALX_METADATA_FACTORY_GUID_KEY = "factory_guid"
    PYQALX_METADATA_PLAN_BOT_NAME_KEY = "plan_bot_name"
    PYQALX_METADATA_BUILD_PATH_KEY = "build_path"

    def __init__(
        self,
        session,
        user_profile=UserConfig.default,
        bot_profile=BotConfig.default,
        aws_profile=default_aws_profile,
    ):
        """
        :param session: An instance of :class:`~pyqalx.core.session.QalxSession`
        :type session: ~pyqalx.core.session.QalxSession
        :param user_profile: The user profile to use
        :type user_profile: str
        :param bot_profile: The bot profile to user
        :type bot_profile: str
        :param aws_profile: The AWS profile to use
        :type aws_profile: str
        """
        self.session = session
        self.user_profile = user_profile
        self.bot_profile = bot_profile
        self.aws_profile = aws_profile
        self.__plan = None
        self.__entity = None
        self.__stage = None
        self.__build_path = None

        self.bot_sources = []
        self.bot_items = []
        self._bots = None

    def _msg_user(self, *args, **kwargs):
        _msg_user(log=self.session.log, *args, **kwargs)

    @property
    def build_path(self):
        """
        If there is an entity on the factory then always use the build
        path defined on the entity otherwise use a random uuid.  Caches the
        build directory on the instance to avoid the uuid changing if there
        is no entity

        :return:  The full build_path
        """
        if self.__build_path is None:
            if self.entity is not None:
                build_basename = self.entity.meta[
                    self.PYQALX_METADATA_NAMESPACE_KEY
                ][self.PYQALX_METADATA_BUILD_PATH_KEY]
            else:
                build_basename = f"builds_{uuid.uuid4()}"
            self.__build_path = os.path.join(
                self.session.config["FACTORY_PACK_DIR"], build_basename
            )
        return self.__build_path

    @property
    def entity(self):
        return self.__entity

    @entity.setter
    def entity(self, factory):
        """
        Setter for entity.  Given an instance of
        ~pyqalx.core.entities.factory.Factory will set the entity on the Factory.
        If given a UUID string will get the factory from the API
        :param factory: An instance of ~pyqalx.core.entities.factory.Factory
        or a uuid as a string
        :raises ValueError: if incorrect instance
        """

        if not isinstance(factory, FactoryEntity):
            if self.__entity is not None:
                # Caches the entity to avoid repeated lookups
                return
            try:
                uuid.UUID(str(factory))
            except ValueError:
                raise QalxFactoryError(
                    f"factory not a valid {FactoryEntity} "
                    f"or {uuid.UUID} instance."
                )

            factory = self.session.factory.get(guid=factory)
        self.__entity = factory

    @property
    def plan(self):
        """
        If there is an entity then use the plan on the entity, otherwise use
        the plan that is already set
        :return: The plan as a JSON string
        """
        if self.entity is not None:
            self.plan = self.entity._file_bytes or self.entity.read_file()
        return self.__plan

    @plan.setter
    def plan(self, plan_yaml):
        """
        Loads the plan from yaml and sets the __plan attribute as JSON
        :param plan_yaml: The plan as a yaml string
        """
        self.__plan = FactoryValidateMixin.load_plan(plan_yaml)

    @property
    def stage(self):
        """
        If there is an entity then use the stage on the entity, otherwise use
        the stage that is already set
        :return: The stage for this factory
        """
        if self.entity is not None:
            self.stage = self.entity.stage
        return self.__stage

    @stage.setter
    def stage(self, value):
        self.__stage = value

    @property
    def _plan_factory(self):
        if self.plan is None:
            raise QalxFactoryError("`plan` must be set on the instance")
        return self.plan["factory"]

    @property
    def _plan_debug(self):
        return FactoryPlanDebug(self._plan_factory.get("debug", {}))

    @property
    def _plan_name(self):
        return self._plan_factory["name"]

    @property
    def _plan_bots(self):
        """
        Helper for getting the bots from the plan
        :return: The bots from the plan
        """
        return self._plan_factory["bots"]

    @property
    def _plan_workflows(self):
        """
        Ensures that the workflows for the plan are always returned as a dict
        of QalxTree instances.  This enables easier iteration and searching of
        the workflow.  Handles there being potentiall multiple root nodes
        by creating a list or dict of the root nodes for each workflow.
        :return: A dict of {workflow_name:QalxTree instances}
        """
        tree_workflows = {}
        for workflow_name, structure in self._plan_factory["workflows"].items():
            tree_workflows[workflow_name] = QalxTree({workflow_name: structure})
        return tree_workflows

    @property
    def _plan_stages(self):
        """
        Helper for getting the stages from the plan
        :return: The stages from the plan
        """
        return self._plan_factory["stages"]

    @property
    def _plan_meta(self):
        """
        Helper for getting the meta from the plan
        :return: The meta from the plan
        """
        return self._plan_factory.get("meta", {})

    @property
    def _plan_tags(self):
        """
        Helper for getting the tags from the plan
        :return: The tags from the plan
        """
        return self._plan_factory.get("tags", {})

    @property
    def _plan_stage(self):
        return self._plan_stages[self.entity.stage]

    @property
    def stacks(self):
        """
        Iterator that returns a stack as a Provider instance.  A stack is a
        unique intersection of a provider type->profile_name->region_name.
        A stack could have multiple sectors on it.  The stack can be accessed
        from the provider instance
        :return: An instance of ~BaseProvider for a specific stack.
        """
        try:
            sectors = Sectors(factory=self)
            for _provider, stacks in sectors.items():
                for profile_name, regions in stacks.items():
                    for region_name, stack in regions.items():
                        Provider = _providers[_provider]
                        provider = Provider(
                            factory=self,
                            profile_name=profile_name,
                            region_name=region_name,
                            stack=stack,
                        )
                        yield provider
        except QalxFactoryError:
            # QalxFactoryError: There is no default region configured
            #                   for a provider
            yield None

    @property
    def bots(self):
        """
        :return: A map of {(<bot_guid>, <plan_bot_name>): <bot_entity>}
                for the bots on the factory entity
        """
        if self._bots is None:
            self._bots = {}
            namespace_key = self.PYQALX_METADATA_NAMESPACE_KEY
            plan_bot_name_key = self.PYQALX_METADATA_PLAN_BOT_NAME_KEY
            factory_guid_key = self.PYQALX_METADATA_FACTORY_GUID_KEY
            query = {
                f"metadata.data.{namespace_key}.{factory_guid_key}": self.entity.guid
            }
            bots = self.session.bot.find(query=query)

            for bot_entity in bots.data:
                plan_bot_name = bot_entity.meta[namespace_key][
                    plan_bot_name_key
                ]
                # A single plan_bot_name could spawn multiple bots in the event
                # that the same plan_bot_name is in different sectors.
                self._bots[(bot_entity.guid, plan_bot_name)] = bot_entity
        return self._bots

    def _delete_build_path(self):
        """
        Deletes the entire build path from the filesystem if it exists. If the
        directory is not found, a message is printed out
        """
        self._msg_user(
            msg=f"Deleting data from build path: `{self.build_path}`"
        )

        def remove_readonly(func, path, _):
            """Clear the readonly bit and reattempt the removal"""
            os.chmod(path, stat.S_IWRITE)
            func(path)

        if os.path.exists(self.build_path):
            shutil.rmtree(self.build_path, onerror=remove_readonly)
        else:
            self._msg_user(
                msg=f"`{self.build_path}` not found on this machine.  "
                f"Unable to delete"
            )

    def _delete_items(self):
        """
        Find all the items for the factory entity based on their
        metadata. Iterates through them deletes them from the API. In the
         case of an error from the API during the deleting process, the error
         is logged and printed out.
        """
        namespace_key = self.PYQALX_METADATA_NAMESPACE_KEY
        factory_guid_key = self.PYQALX_METADATA_FACTORY_GUID_KEY
        query = {
            f"metadata.data.{namespace_key}.{factory_guid_key}": self.entity.guid
        }
        items = self.session.item.find(query=query)
        for item in items.data:
            try:
                self.session.item.delete(item)
            except QalxAPIResponseError as exc:
                err_msg = (
                    f"There was an error deleting a qalx item containing "
                    f"factory bot code.  "
                    f"You may have to delete this manually.: {exc}"
                )
                self._msg_user(msg=err_msg, level=logging.ERROR)

    def _delete_factory_build(self):
        """
        Deletes the factory build entity from the API. In the case of an error,
        the error message is printed and logged
        """
        try:
            self.session.factory.delete(self.entity)
        except QalxAPIResponseError as exc:
            err_msg = (
                f"There was an error deleting the factory build entity.  "
                f"You may have to delete this manually.: {exc}"
            )
            self._msg_user(msg=err_msg, level=logging.ERROR)

    def _delete_stacks(self):
        """
        Iterates through all the stacks and deletes the stack from the
        remote instance.
        """
        deleted_stacks = []
        if self.entity.stacks:
            # May not be stacks if there was an error during stack creation
            for provider in self.stacks:
                if not isinstance(provider, LocalProvider):
                    # LocalProviders won't have a stack to delete.
                    for sector_name in provider.stack.keys():
                        stack_id = self.entity.stacks[sector_name]
                        if stack_id not in deleted_stacks:
                            # Multiple sectors could share the same stack.
                            # Ensure that we only call delete once per stack
                            provider.delete_stack(stack_id=stack_id)
                            deleted_stacks.append(stack_id)

    def _delete_resources(self, delete_stacks=True):
        """
        Helper method for tidying up a factory.  Could be called as part of
        demolishing, or due to a failed build
        :param delete_stacks:Should the stacks be deleted
        :type delete_stacks:bool
        """
        if delete_stacks:
            # Should the stacks be deleted?  If the stack creation failed then
            # the stacks will have deleted themselves, so deletion shouldn't
            # be attempted.  If an error occurs after stack creation then
            # the stacks should be deleted
            self._delete_stacks()
        self._delete_items()
        self._delete_factory_build()
        self._delete_build_path()



import concurrent.futures
import logging

from colorama import Fore

from pyqalx.core.errors import QalxFactoryBuildError
from pyqalx.factories.sectors.providers.provider import BaseProvider


class LocalProvider(BaseProvider):
    """
    A provider subclass for bots that run locally.

    Instance attributes

        _error_reasons:  A list of errors for any bots that have failed
    """

    _type = "local"

    FORE = Fore.GREEN
    REGION_NAME_KEY = "region_name"
    PROFILE_NAME_KEY = "profile_name"
    CREATED = "CREATED"
    DELETED = "DELETED"
    DELETE_FAILED = "DELETED"
    ERRORED = "ERRORED"

    def __init__(self, *args, **kwargs):
        super(LocalProvider, self).__init__(*args, **kwargs)
        self._error_reasons = []
        self._deleted = False

    @staticmethod
    def default_profile_name():
        return "local"

    @staticmethod
    def default_region_name(validate, log):
        return "local"

    @property
    def ERRORED_STATUSES(self):
        return [self.ERRORED]

    @property
    def CREATED_STATUS(self):
        return self.CREATED

    @property
    def DELETED_STATUSES(self):
        return [self.DELETED]

    def _install_and_start_bot(self, bot_source, sector):
        """
        Helper for installing and starting the bot
        :param bot_source: An instance of BotSource
        :param sector: The sector that the bot is on.
        :return: The response from `start_bot`
        """
        try:
            bot_source.install_bot()
        except QalxFactoryBuildError as exc:
            # QalxFactoryBuildError: There was a problem installing the bot

            # Make a fake "output" object that has the returncode
            # attribute so `futures.as_completed` can treat a failure here
            # in the same way as a failure to `start_bot`
            output = type("", (object,), {})()
            output.returncode = 1
            error = str(exc)
        else:
            bot_arguments = sector["bots"][bot_source.bot_name]
            # if skip-ini is present on the plan then it's assumed that
            # the user wants to skip-ini - regardless of the value of the key.
            # This mirrors the behaviour of `--skip-ini` on the CLI
            skip_ini = "skip-ini" in bot_arguments.keys()
            # Need to remove `skip-ini` from the bot arguments as it's handled
            # differently in `self.factory.start_bot`
            bot_arguments.pop("skip-ini", None)

            output, error = self.factory.start_bot(
                plan_bot_name=bot_source.bot_name,
                build_path=bot_source.build_path,
                bot_path=bot_source.bot["bot_path"],
                bot_arguments=bot_arguments,
                sector=sector,
                skip_ini=skip_ini,
            )
        return output, error

    def _create_stack(self, *args, **kwargs):
        """
        Starts each bot on the local sector.  If any error during starting then
        stop immediately and return that error so that the manager can
         handle stopping the build

        :return: The stack_id of `success` if starting the bot is successful
        """
        bots = {}

        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            for sector in self.stack.values():
                for bot_source in self.factory.bot_sources:
                    if bot_source.bot_name in sector["bots"]:
                        bots[
                            executor.submit(
                                self._install_and_start_bot,
                                bot_source=bot_source,
                                sector=sector,
                            )
                        ] = bot_source.bot_name
            for future in concurrent.futures.as_completed(bots):
                plan_bot_name = bots[future]
                output, error = future.result()
                if output.returncode is not None:
                    # There was an error starting up this bot.
                    self._msg_user(str(error), level=logging.ERROR)
                    self._error_reasons.append(f"{plan_bot_name}: {error}")
                else:
                    # Local sectors won't have a stack ID.
                    # But it is still required for the manager to monitor the
                    # creation of all stacks.  If one bot starts, but another local
                    # bot fails, the manager will handle shutting down the started
                    # bot.
                    self.stack_id = "success"
        return self.stack_id

    def _describe_stack(self):
        """
        Local bot startup needs to be atomic.  Therefore, if any of them error
        then we should show the whole stack as failed.
        """
        status = self.CREATED

        if self._error_reasons:
            status = self.ERRORED
        if self._deleted or (self._error_reasons and self.stack_id is None):
            # Regardless of whether or not there are errors,
            # if the _deleted flag is set then the stack has been deleted.
            # There may be errors and no stack id if no bots at all were able
            # to start.  In which case, nothing was created so the stack can be
            # marked as "DELETED"
            status = self.DELETED
        return status

    def _delete_stack(self, stack_id):
        """
        The stack needs to be deleted because either a single local bot failed
        to start, or a remote sector failed to create properly.  There could
        be bots running (albeit in a stopped state), so use the same termination
        process as what would be used when demolishing a factory.  The actual
        bot process will then get terminated by the bot itself as
         if `qalx bot-terminate` had been called.  This will most likely be very
         quick as the bots won't have started working on anything.
        """
        self.factory.breadth_first_terminate()
        # Flag so we can update `_describe_stack`
        self._deleted = True
        return

    def _get_error_reason(self):
        """
        A local stack could have multiple bots running on it.  Therefore
        the error reasons needs to be a list of reasons
        """
        return self._error_reasons or ["Unknown Error"]

    def _build_stack(self):
        """
        This method is used for remote sectors to build a template to build
        remote resources.  As such, a local sector does not require it so it
        should just return True
        """
        return True



from pyqalx.core.entities import QalxListEntity
from pyqalx.core.entities.entity import QalxFileEntity, QalxQueueableEntity


class Item(QalxFileEntity, QalxQueueableEntity):
    """An item is the core of qalx.

    They are structured data or a file or some combination of the two. For example:

    >>> from pyqalx import QalxSession
    >>> qalx = QalxSession()
    >>> dims = qalx.item.add(data={"height":5, "width":5}, meta={"shape":"square"})
    >>> steel = qalx.item.add(source="path/to/316_datasheet.pdf",
    ...                       data={"rho":8000, "E":193e9},
    ...                       meta={"library":"materials",
    ...                             "family":"steel",
    ...                             "grade":"316"})

    We can then use the ``find_one`` and ``find`` methods to search for items

    >>> from pyqalx import QalxSession
    >>> qalx = QalxSession()
    >>> steel_316_item = qalx.item.find_one(query={"metadata.data.library": "materials",
    ...                                            "metadata.data.family": "steel",
    ...                                            "metadata.data.grade": "316"})
    >>> steels = qalx.item.find(query={"metadata.data.family": "steel"})
    >>> squares = qalx.item.find(query={"metadata.data.shape": "square"})
    >>> quads = qalx.item.find(query={"$or": [{"metadata.data.shape": "square"},
    ...                                       {"metadata.data.shape": "rectangle"}]})

    We can edit an item once we have retrieved it and save it back to qalx.
    You can either use attribute style getters/setters (my_shape.data.height = 10)
    or key style getters/setters (my_shape['data']['height'] = 10)

    >>> from pyqalx import QalxSession
    >>> qalx = QalxSession()
    >>> my_shape = qalx.item.find_one(query={"data.height": 5, "data.width": 5})
    >>> my_shape.data.height = 10
    >>> my_shape.meta.shape = 'rectangle'
    >>> qalx.item.save(my_shape)
    >>> # If we think that someone else might have edited my_shape we can reload it:
    >>> my_shape = qalx.item.reload(my_shape)
    """

    entity_type = "item"


class ItemAddManyEntity(QalxListEntity):
    """
    The entity that gets returned when the
    :meth:`~pyqalx.core.adapters.item.QalxItem.add_many` method is called
    """

    _data_key = "items"

    @property
    def items_(self):
        """
        Helper to allow easier lookup for items.  A user can continue to do
        `resp['items']` should they choose - but using this property leads to
        neater code.

        Usage: `resp.items_`

        :return: dict
        :raises: KeyError if `items` key doesn't exist on ItemAddManyEntity
        """
        # We have to use a property rather than an instance attribute even though
        # it's slower because using an instance attribute would cause problems
        # if `add_many` has some items that can't be added (as these get removed
        # from the `items` dict after instantition)
        return self[self._data_key]



"""
CLI package that has all the fuctionality directly related to the cli
implemented with click. qalx is the main function object that provides all the
cli related functionality
"""

from qalxcli.cli.qalx import qalx  # noqa

# The rest of the imports are needed to ensure all the decorators from the other
# modules of cli are bound to the qalx object
from qalxcli.cli import bots  # noqa
from qalxcli.cli import factory  # noqa



import json
from importlib import import_module
from json import JSONDecodeError

import click

from pyqalx import Bot, QalxSession
from pyqalx.config import UserConfig, BotConfig
from pyqalx.core.entities import QalxEntity
from pyqalx.factories.mixins import FactoryValidateMixin


class QalxImport(click.ParamType):
    """
    A custom type to handle how the cli does imports.
    Expects a string in the format `my.dotted.module:attr_to_import`.
    """

    name = "import path"

    def __init__(self, expected_class, is_instance=False, *args, **kwargs):
        """
        :param expected_class: The class that this import should be an
        instance/subclass of
        :param is_instance: Should we do an `isinstance` check?  Defaults to
        doing an `issubclass` check
        """
        self.expected_class = expected_class
        self.is_instance = is_instance
        super(QalxImport, self).__init__(*args, **kwargs)

    def convert(self, value, param, ctx):
        msg = "Argument must be in the format `dotted.path.to:variable_name`"
        try:
            module_name, variable_name = value.split(":")
        except ValueError as exc:
            # ValueError: No `:` in string
            msg = f"{msg}.  Error was `{exc}`"
            self.fail(msg)
        try:
            module = import_module(module_name)
        except ImportError as exc:
            # ImportError: Invalid path
            msg = f"{msg}.  Error was `{exc}`"
            self.fail(msg)
        if variable_name not in dir(module):
            self.fail(f"`{variable_name}` not found in `{module_name}`.")
        _import_target = getattr(module, variable_name)

        method = issubclass
        message = "a subclass"

        if self.is_instance:
            method = isinstance
            message = "an instance"
        if not method(_import_target, self.expected_class):
            self.fail(
                f"`{variable_name}` is not {message} "
                f"of `{self.expected_class.__name__}`."
            )
        return _import_target


class QalxFactoryStage(click.ParamType):
    """
    Custom click type for a factory stage. Checks that the stage plan exists in
    a plan
    """

    name = "factory stage"

    def convert(self, value, param, ctx):
        """
        Check the stage exists in the plan. The plan in parsed from the yml file
        path and the stage must exist in the stages of the provided plan
        """
        plan_yml_path = ctx.params.get("plan")
        plan_yml = FactoryValidateMixin.read_plan(plan_yml_path)
        plan = FactoryValidateMixin.load_plan(plan_yml)
        try:
            stage_exists = value in plan["factory"]["stages"]
        except (KeyError, TypeError):
            # Return the value and the validation can handle the rest
            return value
        if stage_exists:
            return value
        else:
            self.fail("The stage does not exist in the plan", param, ctx)


class QalxDict(click.ParamType):
    """
    Custom click type for key-value pair inputs. Allows such pairs to be passed
    as arguments from the CLI with the following syntax format: `<key>=<value>`.
    Also allows a user to pass a complex object using
    a JSON string: `<key>='{"complex": "structure"}'`
    """

    name = "qalx dict"

    def convert(self, value, param, ctx):
        """
        Check that the input is in the format: "<key>:<value>" and return as a
        dictionary
        """
        # Remove all whitespace
        key_val = "".join((i for i in value.split()))
        try:
            key, val = key_val.split("=")
            try:
                # Attempt to handle metadata being passed as a JSON string
                val = json.loads(val)
            except JSONDecodeError:
                # JSONDecodeError - the metadata is not JSON.  Save it as it is
                pass
            return {key: val}
        except ValueError:
            fail_msg = (
                f'Argument "{key_val}" is not in the expected '
                + '"<key>=<value>" format'
            )
            self.fail(fail_msg)


BOT_IMPORT = QalxImport(Bot, is_instance=True)
QALX_SESSION_IMPORT = QalxImport(QalxSession)
USER_CONFIG_IMPORT = QalxImport(UserConfig)
BOT_CONFIG_IMPORT = QalxImport(BotConfig)
QALX_ENTITY_IMPORT = QalxImport(QalxEntity)
QALX_FACTORY_STAGE = QalxFactoryStage()
QALX_DICT = QalxDict()
POSITIVE_INT = click.IntRange(min=1)
FILE_PATH_EXISTING = click.types.Path(exists=True, file_okay=True)
DIR_PATH_EXISTING = click.types.Path(exists=True)
FILE_ENTITY_TYPE = click.types.Choice(("item", "factory"), case_sensitive=False)



from pyqalx.core.adapters.adapter import QalxNamedEntityAdapter
from pyqalx.core.entities import Queue


class QalxQueue(QalxNamedEntityAdapter):
    """
    Provides an interface for accessing the API for
    :class:`~pyqalx.core.entities.queue.Queue` entities
    """

    _entity_class = Queue
    _bot_only_methods = ["get_messages"]

    @property
    def _queue_params(self):
        """
        The configurable parameters for a `Queue`
        :return:
        """
        return {"VisibilityTimeout": self.session.config["MSG_BLACKOUTSECONDS"]}

    def add(self, name, meta=None, **kwargs):
        """
        Queues are created with a name.  This name is stored in the metadata
        of the `Queue` instance

        :param name: The name we want to assign the Queue
        :type name: str
        :param meta: A dictionary of metadata to store
        :type meta: dict
        :param kwargs: Any other kwargs we are setting on the Queue
        :return: A newly created :class:`~pyqalx.core.entities.queue.Queue` instance
        """
        return super(QalxQueue, self).add(
            parameters=self._queue_params, name=name, meta=meta, **kwargs
        )

    def get_messages(self, worker):
        """
        Gets the messages on the `Queue` instance

        :param worker:An instance of :class:`~pyqalx.core.entities.worker.Worker`
                      that called this method

        :return: A list of :class:`~pyqalx.core.entities.queue.QueueMessage` instances
        """
        config = self.session.config
        max_msgs = config["Q_MSGBATCHSIZE"]
        visibility = config["MSG_BLACKOUTSECONDS"]
        waittime = config["MSG_WAITTIMESECONDS"]

        message = worker.queue.get_messages(
            max_num_msg=max_msgs,
            visibility=visibility,
            waittime=waittime,
            worker=worker,
        )
        return message

    def get_by_name(self, name, **kwargs):
        """a single queue by name

        :param name: name of queue
        :type name: str
        :return: :class:`~pyqalx.core.entities.queue.Queue`
        :raises QalxMultipleEntityReturned:
        :raises QalxEntityNotFound:
        """
        return super(QalxQueue, self).get_by_name(name, **kwargs)

    def get_or_create(self, name, meta=None, **kwargs):
        """
        Gets a Queue by the given name or creates it if it doesn't exist

        :param name:
        :type name: str
        :param meta: metadata about the queue
        :type meta: dict
        :return: :class:`~pyqalx.core.entities.queue.Queue`
        """
        return super(QalxQueue, self).get_or_create(name, meta=meta)

    def get(self, guid, *args, **kwargs):
        # extended to create better autodocs
        return super(QalxQueue, self).get(guid, *args, **kwargs)

    def save(self, entity, *args, **kwargs):
        # extended to create better autodocs
        return super(QalxQueue, self).save(entity, *args, **kwargs)

    def find(
        self,
        query=None,
        sort=None,
        skip=0,
        limit=25,
        many=True,
        include_session_tags=True,
        **kwargs
    ):
        # extended to create better autodocs
        return super(QalxQueue, self).find(
            query=query,
            sort=sort,
            skip=skip,
            limit=limit,
            many=many,
            include_session_tags=include_session_tags,
            **kwargs
        )

    def find_one(self, query, **kwargs):
        # extended to create better autodocs
        return super(QalxQueue, self).find_one(query=query, **kwargs)

    def reload(self, entity, *args, **kwargs):
        # extended to create better autodocs
        return super(QalxQueue, self).reload(entity=entity, **kwargs)

    def aggregate(self, aggregate, **kwargs):
        # extended to create better autodocs
        return super(QalxQueue, self).aggregate(aggregate=aggregate, **kwargs)



import logging
import sys
import traceback
from logging.handlers import RotatingFileHandler

from pyqalx.config import BotConfig, UserConfig
from pyqalx.core.errors import QalxError


# The prefix for all pyqalx log names
LOGGER_NAME_PREFIX = "pyqalx"


class QalxRotatingFileHandler(RotatingFileHandler):
    """
    We always force some defaults for a qalx log
    """

    def __init__(self, *args, **kwargs):
        # 10 MB
        kwargs["maxBytes"] = kwargs.get("maxBytes", 10 * 1024 * 1024)
        kwargs["backupCount"] = kwargs.get("backupCount", 10)
        kwargs["mode"] = kwargs.get("mode", "w")
        super(QalxRotatingFileHandler, self).__init__(*args, **kwargs)

    def rotate(self, *args, **kwargs):
        """
        Overridden in order to ignore permission errors.
        This could happen when multiple processes are accessing the same
        log file.  If this happens, rotation is suppressed until the lock
        is dropped.  The log files will exceed the "maxBytes" value but logging
        will continue normally.
        """
        try:
            super(QalxRotatingFileHandler, self).rotate(*args, **kwargs)
        except PermissionError:
            # PermissionError: Multiple processes writing to same file.  Only
            #                  likely on Windows
            pass


class QalxLogFormatter(logging.Formatter):
    """
    Sometimes our log messages contain extra data.  Use this formatter to allow
    all handlers to use the same format while also handling the potential extra
    data
    """

    def __init__(self, style="%", *args, **kwargs):
        kwargs[
            "fmt"
        ] = "%(asctime)s %(name)-12s %(levelname)-8s %(process)d %(message)s"
        super(QalxLogFormatter, self).__init__(*args, **kwargs)
        self.style = style
        self.original_style = self._style

    def format(self, record):
        dict_record = record.__dict__
        self._style = self.original_style
        if (
            "entity_type" in dict_record.keys()
            and "entity_guid" in dict_record.keys()
        ):
            existing_format = self._style._fmt
            # Typically this is added if a user decorated function encounters
            # an unhandled exception
            existing_format += (
                " Entity Type: `%(entity_type)s`"
                " Entity GUID: `%(entity_guid)s`"
            )
            self._style = logging._STYLES[self.style][0](existing_format)
        return super(QalxLogFormatter, self).format(record)


def configure_logging(name, filename, level):
    """
    Configures a log.  Loggers are tied directly to an instance of
    :class:`~pyqalx.core.session.QalxSession` or :class:`~pyqalx.bot.Worker`
    and should be accessed via `instance.log.<level>`.
    All pyqalx loggers are prefixed with `<LOGGER_NAME_PREFIX>`

    :param name: The name of the log
    :type name: str
    :param filename: The full filename for where the log should be saved
    :type filename: str
    :param level: The level of the log
    :type level: str
    :return: logging.getLogger()
    """
    name = f"{LOGGER_NAME_PREFIX}.{name}"
    if isinstance(level, str):
        # convert to proper log level
        level = level.upper()
    _log = logging.getLogger(name)
    if not _log.handlers:
        handler = QalxRotatingFileHandler(filename=filename)
        handler.setFormatter(QalxLogFormatter())
        handler.setLevel(level)
        _log.addHandler(handler)
    _log.setLevel(level)
    return _log


def session_logging_suffix(config):
    """
    Returns the suffix to use for QalxSession logging instances based
    on the config

    :param config: A subclass of :class:`~pyqalx.config.configs.Config`
    :type config: ~pyqalx.config.configs.Config
    :return: The key_prefix split.
             i.e. `user` for `QALX_USER_` and `bot` for `QALX_BOT_`
    """
    return config.key_prefix.split("_")[1].lower()


def get_logger():
    """
    For all the given configured loggers will try and get a user session log
    to log the unhandled exception to.  If there are multiple configured user
    sessions just grab the first one.  If there are no user sessions
    configured (i.e. the error occurred before QalxSession was instantiated)
    then return the root log
    """
    _loggers = [name for name in logging.root.manager.loggerDict]

    # Figures out which log to use.  Will use the `user session` log first
    # as this is most likely to exist, falling back to the `bot session` log,
    # and finally using the root log.
    for config_type in [UserConfig, BotConfig]:
        suffix = session_logging_suffix(config_type)
        loggers = list(
            filter(lambda x: f"{LOGGER_NAME_PREFIX}.{suffix}" in x, _loggers)
        )
        loggers.sort(key=lambda x: -len(x))
        if loggers:
            # A log has been found, stop trying to find one
            break
    logger_name = logging.root.name
    if loggers:
        logger_name = loggers[0]
    return logging.getLogger(logger_name)


def exception_hook(exc_type, exc_value, exc_traceback, extra=None, log=None):
    """
    We want to log any unhandled exception that occurs in pyqalx.  This
    includes `Qalx*` errors (i.e. because a user has misconfigured something)
    and also if they have made an error in one of their decorated functions
    (i.e. @bot.process).  This catches all errors and logs them and also
    prints out the traceback
    """
    # If no log (i.e. it's an error in the normal codebase) then figure out
    # which log to use
    log = log or get_logger()

    if issubclass(exc_type, QalxError):
        # Only handle `Qalx*` Errors.  We don't want to consume anything the
        # user has done wrong.  This will however handle issues with decorated
        # bot functions as these are caught in ~bot.Bot and handled as a
        # QalxError.  (i.e. if the user does `int('a')` in a bot function the
        # exception will be logged and raised here.)
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return

        log.error(
            msg="Qalx Exception:",
            exc_info=(exc_type, exc_value, exc_traceback),
            extra=extra,
        )

        traceback.print_exception(
            exc_type, exc_value, exc_traceback, chain=False
        )
    else:
        log.error(
            msg="Unhandled Exception:",
            exc_info=(exc_type, exc_value, exc_traceback),
            extra=extra,
        )

        # This is an exception elsewhere in the users codebase outside of qalx
        # or any bot function. We don't handle it.  Just reraise it for
        # the user to fix
        raise exc_type(exc_value)


sys.excepthook = exception_hook



import itertools
import os

import click
from click import ClickException
from tabulate import tabulate

from pyqalx.core.encryption import QalxEncrypt
from pyqalx.core.entities import QalxListEntity
from pyqalx.core.entities.bot import Bot
from pyqalx.core.entities.factory import Factory
from pyqalx.core.signals import QalxSignal


def generate_key_file():
    """
    Generate an encryption key file and return the path to the new file if this
    has been successful
    """
    while True:
        key_path = click.prompt(
            "Where do you want to save the key file?",
            default=QalxEncrypt.DEFAULT_KEY_PATH,
        )
        if os.path.exists(key_path):
            click.echo(
                f"\nThe provided path: {key_path} exists. "
                + "Please provide a non-existing path\n"
            )
        else:
            break
    try:
        QalxEncrypt.save_key(key_path)
    except (OSError, PermissionError) as exc:
        message = (
            "There was an error saving the encryption key file:\n\n" + f"{exc}"
        )
        raise ClickException(message)
    else:
        return key_path


class QalxCLITabulation:
    """
    A helper class for building a table of entities to display to the user
    """

    prompt_text = None

    def __init__(self, qalx_session, name, entity_class, tabulate_single=False):
        """
        :param qalx_session: An instance of QalxSession
        :param name: The name of the entity that we are looking for
        :param entity_class: The entity class that is being tabulated
        :param tabulate_single: Whether we should tabulate if a single
        result is found or just return the entity
        """

        self.qalx_session = qalx_session
        self.name = name
        self.tabulate_single = tabulate_single
        self.entity_class = entity_class
        self.entity_type = entity_class.entity_type
        self.next_table_index = 1

        # Get the entities for the prompt
        self.entities_for_prompt = self.get_entities_for_prompt()

    def query_kwargs(self):
        """
        By default there are no additional kwargs passed through to the query
        """
        return {}

    def get_entities_for_prompt(self):
        """
        Queries the API and returns the entities for the prompt based on
        `self.build_query()` and `self.query_kwargs`
        """
        query = self.build_query()
        entities = getattr(self.qalx_session, self.entity_type).find(
            query=query, **self.query_kwargs()
        )
        return entities

    def _update_entities_for_prompt(self, updated_entities):
        """
        Update the entities for prompt. This is needed in the case of paginated
        results from a query.
        """
        self.entities_for_prompt = updated_entities

    def build_query(self):
        """
        Builds the query that will be used to query the api
        :return: A dict of the query, by default just the name will be queried
        """
        return {"name": self.name}

    def build_no_entities_message(self):
        """
        The message to be displayed if no entities are found matching the query
        """
        return f"No {self.entity_class.pluralise()} found called `{self.name}`"

    @staticmethod
    def build_row(index, entity):  # pragma: no cover
        """
        A specific row on a table
        :param index: The index of this row
        :param entity: The entity instance
        :return: A list of data for the entity for a single row
        """
        raise Exception("Must be implemented by subclass")

    @staticmethod
    def build_headers():  # pragma: no cover
        """
        The headers for this table
        :return: A list of strings for the table
        """
        raise Exception("Must be implemented by subclass")

    def tabulate(self):
        """
        For the given `self.entities_for_prompt` will build a table using
        tabulate and display this to the user via `click.echo`
        """
        table = []

        # Massage the remaining entities into a nice format for presentation
        entities_for_prompt = self.entities_for_prompt.data
        for index, entity in enumerate(
            entities_for_prompt, start=self.next_table_index
        ):
            table.append(self.build_row(index, entity))
            self.next_table_index += 1

        # We then display the entities to the user in a list,
        # and potentially prompt them for an index for the specific one they
        # wish to interact with
        headers = self.build_headers()
        # Display a table of entities
        click.echo(tabulate(table, headers, maxcolwidths=20))
        return table

    def get_entity_or_display(self):
        """
        Will either return the entity or display a table of entities
        depending on the number of entities found and the value
        of `self.tabulate_single`.
        If no entities are found always return None
        """
        entities_for_prompt = self.entities_for_prompt.data
        if len(entities_for_prompt) == 1:
            # Only a single entity found matching the name and with active
            # workers.
            # Just return it without showing the user the table of entities
            if self.tabulate_single:
                self.tabulate()
            else:
                return entities_for_prompt[0]
        elif not len(entities_for_prompt):
            # No entities found.  Nothing to do!
            click.echo(self.build_no_entities_message())
            return
        else:
            # Many entities found matching the query.
            # Create a table to allow the user to pick which they want
            table = self.tabulate()

            if (
                self.prompt_text
                or self.entities_for_prompt.query.next is not None
            ):
                while True:
                    # Message the user they can show next page in case of paginated
                    # results
                    has_next_page = (
                        self.entities_for_prompt.query.next is not None
                    )
                    prompt_message = (
                        "(Press ENTER to show the next page)\n\n"
                        if has_next_page
                        else ""
                    )
                    if self.prompt_text:
                        prompt_message += (
                            f"Please choose a {self.entity_type} index "
                            + f"to {self.prompt_text}"
                        )
                    if not prompt_message:
                        return
                    inp = click.prompt(prompt_message, default="")

                    if has_next_page and inp == "":
                        # If the user presses ENTER then we show the next page
                        # and append the results to the existing variables
                        response = getattr(
                            self.qalx_session, self.entity_type
                        )._process_api_request(
                            "get", self.entities_for_prompt.query.next
                        )
                        entities = QalxListEntity(
                            response, child=self.entity_class
                        )

                        self._update_entities_for_prompt(entities)

                        entities_for_prompt += self.entities_for_prompt.data
                        table += self.tabulate()
                    else:
                        # Case where they have provided an index. This needs to
                        # be in the range of the table that was built
                        if self.prompt_text:
                            try:
                                index = int(inp)
                            except ValueError:
                                pass
                            else:
                                if 0 < index <= len(table):
                                    # Get the entity from the index (- 1 because
                                    # the index the user chooses starts at 1)
                                    return entities_for_prompt[index - 1]
                        else:
                            return


class BotTabulation(QalxCLITabulation):
    def __init__(self, *args, **kwargs):
        super(BotTabulation, self).__init__(entity_class=Bot, *args, **kwargs)

    @staticmethod
    def build_headers():
        return [
            "Index",
            "Name",
            "State",
            "Platform",
            "Node",
            "No. Workers \n(Active/Total)",
            "Created On \n(UTC)",
            "Created By",
        ]

    @staticmethod
    def build_row(index, entity):
        created_on = entity.info.created.on.strftime("%d/%m/%Y %H:%M:%S")
        num_active_workers = len(
            [worker for worker in entity.workers if worker.state == "active"]
        )
        row = [
            index,
            entity.name,
            entity.state,
            entity.host.get("platform"),
            entity.host.get("node"),
            f"{num_active_workers}/{len(entity.workers)}",
            created_on,
            entity.info.created.by.email,
        ]
        return row


class TerminateBotTabulation(BotTabulation):
    prompt_text = "terminate"

    def build_query(self):
        query = super(TerminateBotTabulation, self).build_query()
        query["workers"] = {"$exists": True, "$not": {"$size": 0}}
        return query

    def build_no_entities_message(self):
        message = super(
            TerminateBotTabulation, self
        ).build_no_entities_message()
        return f"{message} that have workers that are not already terminated"

    @staticmethod
    def _filterfalse_signal(_worker):
        """
        itertools.filterfalse is used to exclude workers if a specific signal
        IS FALSE.
        Returns the status of this workers terminate signal.  If this signal
        is True (i.e. the worker is terminated) then it WON'T get returned
        to the user
        """
        return QalxSignal(_worker).terminate

    def _filter_entities(self, bots_for_prompt):
        """
        Filter the bots to only show those that have workers that aren't
        terminated
        """

        def _has_workers_with_specific_signal(_bot):
            # For each bot, we only return it if any of the
            # workers ARE NOT on a status of `self._filterfalse_signal`.
            # As a separate function to avoid nested lambdas
            return list(
                itertools.filterfalse(self._filterfalse_signal, _bot.workers)
            )

        # We then filter the bots to only be those that have any workers that
        # don't have a self._filterfalse signal.
        bots = list(
            filter(_has_workers_with_specific_signal, bots_for_prompt.data)
        )
        bots_for_prompt.data = bots
        return bots_for_prompt

    def get_entities_for_prompt(self):
        """
        For the termination prompt, only show bots that exist with workers
        that aren't terminated.
        """
        entities_for_prompt = super(
            TerminateBotTabulation, self
        ).get_entities_for_prompt()
        return self._filter_entities(entities_for_prompt)

    def _update_entities_for_prompt(self, updated_entities):
        self.entities_for_prompt = self._filter_entities(updated_entities)


class StopBotTabulation(TerminateBotTabulation):
    prompt_text = "stop"

    def build_no_entities_message(self):
        message = super(
            TerminateBotTabulation, self
        ).build_no_entities_message()
        return (
            f"{message} that have workers that are not already stopped "
            f"or terminated"
        )

    @staticmethod
    def _filterfalse_signal(_worker):
        """
        itertools.filterfalse is used to exclude workers if a specific signal
        IS FALSE.
        :return: If the workers stop signal is True (i.e. the worker is stopped)
        then the worker WON'T get returned to the user OR if the workers
        terminate signal is True (i.e. the worker is terminated)
         then the worker WON'T get returned to the user.
         Therefore, this will only return bots with non terminated workers that
         are also not stopped
        """
        signal = QalxSignal(_worker)
        return signal.terminate is True or signal.stop is True


class ResumeBotTabulation(TerminateBotTabulation):
    prompt_text = "resume"

    def build_no_entities_message(self):
        message = super(
            TerminateBotTabulation, self
        ).build_no_entities_message()
        return f"{message} that have workers that can be resumed"

    @staticmethod
    def _filterfalse_signal(_worker):
        """
        itertools.filterfalse is used to exclude workers if a specific signal
        IS FALSE.
        :return: if the workers stop signal is False (i.e. the worker is not stopped)
        then the worker WON'T get returned to the user OR if the workers terminate
        signal is True (i.e. the worker is terminated) then the worker WON'T get
        returned to the user.
        Therefore, this will only return bots with non terminated workers that
        are also stopped
        """
        signal = QalxSignal(_worker)
        return signal.terminate is True or signal.stop is False


class FactoryTabulation(QalxCLITabulation):
    def __init__(self, *args, **kwargs):
        super(FactoryTabulation, self).__init__(
            entity_class=Factory, *args, **kwargs
        )

    @staticmethod
    def build_headers():
        return [
            "Index",
            "Name",
            "Stage",
            "Status",
            "Created On (UTC)",
            "Created By",
        ]

    @staticmethod
    def build_row(index, entity):
        created_on = entity.info.created.on.strftime("%d/%m/%Y %H:%M:%S")
        row = [
            index,
            entity.name,
            entity.stage,
            entity.status,
            created_on,
            entity.info.created.by.email,
        ]
        return row


class DemolishFactoryTabulation(FactoryTabulation):
    prompt_text = "demolish"



import os
import sys
from functools import partial
from multiprocessing import Process
from time import sleep
from collections import OrderedDict

# We may need below, we may not. What can I say? Dill waters run deep.
import dill

from pyqalx import QalxBot
from pyqalx.config import BotConfig, UserConfig
from pyqalx.core.adapters import QalxQueue, QalxWorker
from pyqalx.core.enums import ValidStates
from pyqalx.core.session import QalxSession
from pyqalx.core.errors import (
    QalxBotInitialisationFailed,
    QalxStepFunctionNotDefined,
    QalxMultipleEntityReturned,
    QalxEntityNotFound,
    QalxAPIResponseError,
)
from pyqalx.core.errors import QalxError
from pyqalx.core.log import exception_hook, configure_logging
from pyqalx.core.signals import QalxSignal
from pyqalx.factories.factory import Factory

dill.settings["recurse"] = True


def do_nowt(*args, **kwargs):
    pass


def do_nowt_right(*args, **kwargs):
    return True


def cannot_be_undefined(*args, **kwargs):
    """
    Raises an exception if a step function cannot be left undefined.
    Handled here to avoid a user trying to call the specific step function
    directly.
    """
    raise QalxStepFunctionNotDefined(
        f"`{kwargs['step_name']}` function must be defined either "
        f"as a decorator (`@mybot.{kwargs['step_name']}`) or by "
        f"overriding the function entirely "
        f"(`mybot.{kwargs['step_name']}_function = <{kwargs['step_name']}_function>`)"
    )


class QalxJob:
    """a Job to process

    :param worker_process: The instance of the worker that is running this job
    :type worker_process: ~pyqalx.bot.Worker
    """

    __slots__ = (
        "worker_adapter",
        "entity_adapter",
        "entity",
        "session",
        "state",
        "options",
        "progress",
        "warnings",
        "errors",
        "_current_step",
        "_process",
        "_worker_entity",
        "_bot_entity",
        "_queue_message",
        "_step_results",
        "context",
        "do_workflow",
    )

    def __init__(self, worker_process):
        self._process = worker_process
        self.worker_adapter = worker_process.worker_adapter
        self._bot_entity = worker_process.bot_entity
        self._worker_entity = worker_process.api_entity
        self.session = worker_process.session

        self._current_step = "class init"
        # We need to preserve the order of step results and although in Python
        # >= 3.6 this is implemented in the dict class, it cannot be relied upon
        # https://docs.python.org/3.6/whatsnew/3.6.html#whatsnew36-compactdict
        self._step_results = OrderedDict()
        self.context = dict()
        self.do_workflow = True

    @property
    def log(self):
        """
        An instance of the workers log to enable logging from step functions
        """
        return self._process.log

    @property
    def e(self):
        """shortcut to QalxJob().entity"""
        return self.entity

    @property
    def s(self):
        """shortcut to QalxJob().session"""
        return self.session

    @property
    def queue(self):
        """shortcut to the queue instance on the bot entity"""
        return self.session.queue.get_by_name(
            self._bot_entity.config.queue_name
        )

    def stop_processing(self):
        """
        Stop the worker from processing.  Will cause the worker to wait until
        it receives a `resume` signal
        """
        self.worker_adapter.stop(
            entity=self._worker_entity, bot_entity=self._bot_entity
        )

    def resume_processing(self):
        """
        Resume the worker.
        """
        self.worker_adapter.resume(
            entity=self._worker_entity, bot_entity=self._bot_entity
        )

    def terminate(self, warm=True, requeue_job=False):
        """
        The job has been told to terminate the worker

        :param warm: If True will completely finish running through the
                     processes for the current job.  If False will exit as soon
                     as the current process function is complete
        :type warm: bool
        :param requeue_job: Should the job be requeued if we are terminating
                            with `warm=False`
        :type requeue_job: bool
        """
        self.worker_adapter.terminate(
            entity=self._worker_entity,
            bot_entity=self._bot_entity,
            warm=warm,
            requeue_job=requeue_job,
        )

    def publish_status(self, status):
        """update the status of the worker

        :param status: short message about the worker status
        :type status: str
        """
        self.worker_adapter.update_status(
            entity=self._worker_entity,
            bot_entity=self._bot_entity,
            status=status,
        )

    def add_step_result(self, success=True):
        """indicate success or failure of the step function

        :param success: was the step successful or not
        :type success: bool
        """
        self._step_results[self._current_step] = success

    @property
    def last_step_result(self):
        """
        The success flag for the previous step result (or None if no previous
        steps) for this job
        """
        if self._step_results:
            return self._step_results[list(self._step_results)[-1]]
        else:
            return None

    def get_entity(self, entity_type, entity_guid):
        """
        get an entity

        :param entity_type: the type of entity e.g. item, set, group
        :type entity_type: str
        :param entity_guid: the guid of the entity
        :type entity_guid: str
        :return:    a child of :class:`~pyqalx.core.entities.entity.QalxEntity`
                    if the entity is found on the API, otherwise None
        """
        adapter = getattr(self.session, entity_type)
        self.entity_adapter = adapter
        try:
            self.entity = adapter.get(entity_guid, config=self.session.config)
        except QalxAPIResponseError:
            self.log.error(
                f"The {entity_type} with guid {entity_guid} was not found"
                + " on qalx. This probably means that it was manually "
                + f"deleted. The {entity_type} was not processed and the "
                + "message will be deleted from the queue."
            )
            self.delete_message()
            self.entity = None
        return self.entity

    def add_item_data(self, item_key, data, meta=None):
        """
        helper method to add item data to the set

        :param item_key: key on the set to add the item as
        :type item_key: str
        :param data: data to add
        :type data: dict
        :param meta: metadata to add
        :type meta: dict

        :return: The specific :class:`~pyqalx.core.entities.item.Item` instance
                 that has been added
        """
        if not (self.entity.entity_type == "set"):
            raise TypeError(
                "Method add_item_data only works on jobs with a set entity."
            )
        self.entity.items_[item_key] = self.session.item.add(
            data=data, meta=meta
        )
        self.save_entity()
        return self.entity.items_[item_key]

    def save_entity(self):
        """
        Updates the API representation of this entity with the data
        currently stored in the job.

        :return: a child of :class:`~pyqalx.core.entities.entity.QalxEntity` of
                 the correct type based on the type of `self.entity`
        """
        # Always ensure that the entity on the adapter is fully up to date
        # with the entity on the job.
        self.entity_adapter.save(entity=self.entity)
        return self.entity

    def reload_entity(self):
        """
        Reloads `self.entity` from the API

        :return: a child of :class:`~pyqalx.core.entities.entity.QalxEntity` of
                 the correct type based on the type of `self.entity`
        """
        # Always ensure that the entity on the adapter is fully up to date
        # with the entity on the job.
        self.entity = self.entity_adapter.reload(entity=self.entity)
        return self.entity

    def delete_message(self):
        """
        Helper method for deleting a message from a Queue
        """
        return self._process._delete_message()


class Bot:
    """
    Bots allow automation of tasks.  A bot reads messages from a
    :class:`~pyqalx.core.entities.queue.Queue` and processes the response
    through user configured `step_functions`.
    """

    __slots__ = (
        "name",
        "queue_name",
        "initialisation_function",
        "begin_function",
        "preload_function",
        "onload_function",
        "preprocess_function",
        "process_function",
        "precompletion_function",
        "postprocess_function",
        "onwait_function",
        "ontermination_function",
        "bot_entity",
        "bot_adapter",
        "qalx_session_class",
        "user_config_class",
        "bot_config_class",
        "entity_classes",
        "workflow",
    )

    def __init__(self, bot_name):
        """
        :param bot_name: The unique name for this bot
        :type bot_name: str
        """
        self.name = bot_name

        # bot lifecycle functions
        self.initialisation_function = do_nowt_right
        self.begin_function = do_nowt
        self.preload_function = do_nowt
        self.onload_function = do_nowt
        self.preprocess_function = do_nowt

        # Use `partial` so we can ensure the name of the function always goes
        # to the exception.
        self.process_function = partial(
            cannot_be_undefined, step_name=self.process.__name__
        )
        self.precompletion_function = do_nowt
        self.postprocess_function = do_nowt
        self.onwait_function = do_nowt
        self.ontermination_function = do_nowt

    def _collect_steps(self):
        steps = {}

        for fn in self.__dir__():
            if fn.endswith("_function"):
                fn_attr = getattr(self, fn)
                if fn_attr not in [do_nowt, do_nowt_right]:
                    if getattr(fn_attr, "func", None) == cannot_be_undefined:
                        # This function must be defined, call it to raise
                        # an exception pre fork.  Only functions using
                        # `partial` will have the `func` attr
                        fn_attr()
                    steps[fn] = dill.dumps(fn_attr)  # serialise functions
        return steps

    def start(
        self,
        queue_name,
        processes=1,
        qalx_session_class=None,
        user_config_class=None,
        bot_config_class=None,
        entity_classes=None,
        user_profile_name=UserConfig.default,
        bot_profile_name=BotConfig.default,
        skip_ini=False,
        stop=False,
        meta=None,
        workflow=None,
    ):
        """
        Starts the bot with `processes` number of :class:`~pyqalx.bot.Worker`
        instances.  Each worker will be able to read a message from the queue
        in parallel

        :param queue_name: The unique name for the queue
        :type queue_name: str
        :param qalx_session_class: The :class:`~pyqalx.core.session.QalxSession`
                                   class that this Bot and Workers should use
        :type qalx_session_class: pyqalx.core.session.QalxSession
        :param user_config_class: The :class:`~pyqalx.config.configs.UserConfig`
                                  class that will be used when creating the Bot
        :type user_config_class: pyqalx.config.configs.UserConfig
        :param bot_config_class: The :class:`~pyqalx.config.configs.BotConfig`
                                 class that will be used when using the bot session
        :type bot_config_class: pyqalx.config.configs.BotConfig
        :param entity_classes: A list of
                               :class:`~pyqalx.core.entities.entity.QalxEntity`
                               subclasses that you want this Bot to use once
                               it is started
        :type entity_classes: list
        :param user_profile_name: The user profile name that will be used when
                                  creating the Bot
        :type user_profile_name: str
        :param bot_profile_name: The bot profile name that will be used when
                                 using the bot session
        :type bot_profile_name: str
        :param skip_ini: Should reading from the `ini` file be skipped when
                         instantiating the user and bot sessions
        :type skip_ini: bool
        :param processes: The number of :class:`~pyqalx.bot.Worker` processes
                          this bot should spawn
        :type processes: int
        :param stop:    Should the bot be stopped right after startup. This is
                        essential functionality during factory building
        :type  stop:    bool
        :param meta:        Metadata to be saved on the bot. Optional parameter
        :type  meta:        dict
        :param workflow:    An iterable of other bots that defines the
                            interaction of the bot with other bots in the
                            context of a workflow
        :type  workflow:    list
        """

        self.queue_name = queue_name
        # This is needed for starting outside the CLI
        workflow = workflow or []
        if not isinstance(workflow, list) or not all(
            [isinstance(_, str) for _ in workflow]
        ):
            raise QalxError(
                "Workflow argument must be an iterable containing strings"
            )
        self.workflow = workflow

        if entity_classes is None:
            entity_classes = []

        if qalx_session_class is None:
            qalx_session_class = QalxSession

        user_session = qalx_session_class(
            profile_name=user_profile_name,
            config_class=user_config_class,
            skip_ini=skip_ini,
        )

        if bot_config_class is None:
            bot_config_class = BotConfig

        # This configures logging to use the config settings specified
        # on the bot
        bot_session = qalx_session_class(
            profile_name=bot_profile_name,
            config_class=bot_config_class,
            skip_ini=skip_ini,
        )
        queue = user_session.queue.get_or_create(name=self.queue_name)
        # Add the queue_name - this makes it easier to find bots looking at
        # specific queues. Also add the `_region` in case the queue gets
        # reloaded by the worker (which will be using the `bot_session`)
        bot_session.config.update(
            {"queue_name": self.queue_name, "REGION": queue._region}
        )
        # Log all the user and bot session config values for debugging
        for session, title in zip((user_session, bot_session), ("user", "bot")):
            session.log.debug(f"-vvv- Config values for {title} session -vvv-")
            for k, v in session.config.items():
                if k != "TOKEN":
                    session.log.debug(f"{k}: {v}")
            session.log.debug(f"-^^^- Config values for {title} session -^^^-")
        self.bot_entity = user_session.bot.add(
            name=self.name, config=bot_session.config, meta=meta
        )

        self.bot_adapter = bot_session.bot

        self._register_entities(entity_classes)

        if self.initialisation_function:
            init_result = self.initialisation_function(user_session)
            bot_session.log.debug(
                f"Initialisation function returns `{init_result}`"
            )
            if not init_result:
                raise QalxBotInitialisationFailed(
                    "The initialisation function returned a False-like value."
                )
        steps = self._collect_steps()

        # Update the token on the bot_session to ensure all subsequent
        # requests use the token from the bot
        token = self.bot_entity.info.credentials.token
        bot_session.rest_api.token = token

        self.bot_entity = self.bot_adapter.replace_workers(
            bot_entity=self.bot_entity, workers=processes
        )
        if stop is True:
            self.bot_adapter.stop(self.bot_entity)
        workers = []
        # We use the same session and adapter for all workers

        worker_adapter = QalxWorker(session=self.bot_adapter.session)

        for worker in range(processes):
            workers.append(
                Worker(
                    steps,
                    queue,
                    self.bot_entity,
                    self.bot_entity.workers[worker],
                    worker_adapter,
                    index=worker,
                    workflow=self.workflow,
                )
            )
            bot_session.log.debug(
                f"Created worker `{worker}` to read from `{self.queue_name}`"
            )
        for p in workers:
            p.start()
        for p in workers:
            p.join()
        bot_session.log.debug("Workers joined.")

    def _register_entities(self, entity_classes):
        """
        Registers `self.entity_classes` for this Bots session
        """
        for entity_class in entity_classes:
            self.bot_adapter.session.register(entity_class)

    def initialisation(self, func):
        """decorator to to register the initialisation

        `initialisation` is executed with a QalxSession instance
        authenticated againt the `user_config_class` __init__ argument
        :param func: function to be executed
        """
        self.initialisation_function = func

    def begin(self, func):
        """decorator to to register the begin

        `begin` is executed with a QalxJob instance

        :param func: function to be executed
        """
        self.begin_function = func

    def preload(self, func):
        """decorator executed before entity is loaded from qalx.

        `preload` is executed with a QalxJob instance

        :param func: function to be executed
        """
        self.preload_function = func

    def onload(self, func):
        """decorator to register the onload

        `onload` is executed with a QalxJob instance

        :param func: function to be executed
        """
        self.onload_function = func

    def preprocess(self, func):
        """decorator to register preprocess function

        `preprocess` is executed with a QalxJob instance

        :param func: function to register
        """
        self.preprocess_function = func

    def process(self, func):
        """decorator to register process function

        `process` is executed with a QalxJob instance

        :param func: function to register
        """
        self.process_function = func

    def precompletion(self, func):
        """decorator to register precompletion function

        `precompletion` is executed with a QalxJob instance

        :param func: function to register
        """
        self.precompletion_function = func

    def postprocess(self, func):
        """decorator to register postprocess function

        `postprocess` is executed with a QalxJob instance

        :param func: function to register
        """
        self.postprocess_function = func

    def onwait(self, func):
        """decorator to register onwait function

        `onwait` is executed with a QalxJob instance

        :param func: function to register
        """
        self.onwait_function = func

    def ontermination(self, func):
        """decorator to register ontermination function

        `ontermination` is executed with a QalxJob instance

        :param func: function to register
        """
        self.ontermination_function = func


class Worker(Process):
    """

    exit codes (start at 72590):

        - 72590 base exit code
        - 72591 warm exit
        - 72592 cold exit
        - 72593 config triggered (e.g. KILL_AFTER)
    """

    def __init__(
        self,
        steps,
        queue,
        bot_entity,
        api_entity,
        worker_adapter: QalxWorker,
        index,
        workflow,
        **kwargs,
    ):
        super(Worker, self).__init__(**kwargs)
        self._steps = steps
        self.queue = queue
        self.bot_entity = bot_entity
        self.api_entity = api_entity
        self.worker_adapter = worker_adapter
        self.session = self.worker_adapter.session
        # Reinitialise the session due to mulitprocessing
        # https://github.com/psf/requests/issues/4323
        self.session.rest_api._initialise_session()

        self.log = None
        self.queue_adapter = QalxQueue(session=self.session)
        self.bot_adapter = QalxBot(session=self.session)
        self.stop = False
        self.config = self.session.config
        self.index = index
        self.workflow = workflow
        self.worker_signal = None

    def configure_logging(self):
        """
        Configures the log for the Worker. Each Worker logs to its own
        file. This prevents issues with multiple processes trying to write
        to the same file
        """
        directory = self.config["WORKER_LOG_FILE_DIR"]
        level = self.config["LOGGING_LEVEL"]
        filename = os.path.join(
            directory, f".worker_{self.bot_entity.name}.{self.index}.log"
        )
        self.log = configure_logging(f"worker.{self.index}", filename, level)

    def _reload_signal(self):
        """
        Helper method for reloading the signal for this worker
        """
        self.worker_signal = self.worker_adapter.get_signal(
            self.api_entity, bot_entity=self.bot_entity
        )

    def _check_worker_signal(self):
        """
        Reload the worker signal and either terminates the worker or sets the
        stop flag so that the worker can stop processing the next time
        `self._continue_processing` is checked
        """
        # The first time the bot starts the signal won't be set yet unless the
        # bot itself was started with `stop=True` (i.e. when building factories)
        stopped_status = getattr(self.worker_signal, "stop", False)

        # Once we've reloaded the signal we can decide what needs to be done
        # with the bot
        self._reload_signal()

        if self.worker_signal.terminate_warm:
            self._do_exit(**QalxSignal._warm_exit_update())
        else:
            if self.worker_signal.stop != stopped_status:
                # The worker stop signal has changed state - log it so the user
                # knows what's going on
                update = QalxSignal._stop_or_resume_update(
                    stop=self.worker_signal.stop
                )
                self.log.info(f'[{update["code"]}] {update["update"]}')
            self.stop = self.worker_signal.stop

    def _do_exit(self, update, code=QalxSignal.BASE_EXIT_CODE):
        self.job._current_step = "OnTerminate"
        onterm = self.steps.get("ontermination_function")
        if onterm:
            onterm(self.job)
        self.job.publish_status(update)
        self.log.info(f"[{code}] {update}")
        sys.exit(code)

    def _check_cold_exit(self, reload_signal=True):
        """
        Reloads the worker signal and checks whether we should do a cold
        termination
        """
        if reload_signal:
            self._reload_signal()

        if self.worker_signal.terminate_cold:
            # The worker has been asked to cold exit for some reason.
            # Check if the job needs to go back on the queue or if we can
            # delete it
            if self.job._queue_message:
                if self.worker_signal.requeue_job:
                    self.job._queue_message._stop_heartbeat()
                else:
                    self._delete_message()
            self._do_exit(**QalxSignal._cold_exit_update())

    def _check_warm_exit(self):
        """
        Reloads the worker signal and checks whether we should do a warm
        termination
        """
        self._reload_signal()

        if self.worker_signal.terminate_warm:
            # The worker has been asked to warm exit for some reason.
            # The job will have finished and the message should have been
            # deleted.
            self._do_exit(**QalxSignal._warm_exit_update())

    def _workflow(self):
        """
        Workflow step function. Used to handle passing a job to multiple bots
        once all step functions have been completed on the worker
        """
        self.log.debug("Attempting worklow")
        factory_namespace_key = Factory.PYQALX_METADATA_NAMESPACE_KEY
        factory_guid_key = Factory.PYQALX_METADATA_FACTORY_GUID_KEY
        factory_plan_bot_name_key = Factory.PYQALX_METADATA_PLAN_BOT_NAME_KEY
        factory_guid = self.bot_entity.meta.get(factory_namespace_key, {}).get(
            factory_guid_key, None
        )
        for bot_name in self.workflow:
            self.log.debug(
                f"Workflow found." f"Next bot in workflow: `{bot_name}`"
            )
            # By default, just try and find the bot based on the name
            query = {"name": bot_name}
            if factory_guid is not None:
                # If it's a factory bot then we need to use the plan_bot_name
                # and the factory_guid, as the actual bot_name won't be known
                # when the factory was built
                pyqalx_query = f"metadata.data.{factory_namespace_key}"
                query = {
                    "metadata.data._class": "pyqalx.factory",
                    f"{pyqalx_query}.{factory_plan_bot_name_key}": bot_name,
                    f"{pyqalx_query}.{factory_guid_key}": factory_guid,
                }
            try:
                bot = self.session.bot.find_one(query=query)
            except (QalxMultipleEntityReturned, QalxEntityNotFound) as exc:
                warn_msg = (
                    f"There was an error retrieving one bot "
                    f"for `{bot_name}` in the defined workflow. "
                    f"Query was `{query}`: `{exc}`"
                )
                self.log.warning(warn_msg)
                continue
            queue = self.session.queue.get_by_name(name=bot.config.queue_name)
            self.log.debug(
                f"Submitting `{self.job.entity}` "
                f"to `{queue.name}` on workflow bot"
                f" name `{bot_name}`"
            )
            self.job.entity.add_to_queue(self.job.entity, queue)

    def _delete_message(self):
        """
        Helper function for stopping any heartbeat on the job and
        deleting the message
        """
        if getattr(self.job, "_queue_message", None):
            self.queue.delete_message(self.job._queue_message)

    def _wait_with_spreading(self, spread_factor):
        """
        We want to wait for a while before trying something again and wait
        longer and longer every time up to a maximum. This is to help limit the
        request count to qalx while staying responsive.

        :return: the next spreading factor to use
        """

        time_to_wait = self.config["Q_WAITTIMESECONDS"]
        max_spreading = self.config["Q_WAITTIMEMAXSPREADING"]
        self.job._queue_message = None
        self.job._current_step = "OnWait"
        if self.steps.get("onwait_function") is not None:
            self.steps["onwait_function"](self.job)
        total_wait_seconds = time_to_wait * spread_factor
        sleep(total_wait_seconds)
        self.log.debug(f"wait for {total_wait_seconds}s")
        self._check_worker_signal()
        self._check_cold_exit(reload_signal=False)  # we reloaded in prev line
        next_spread = spread_factor + 1
        # never spread beyond a max factor
        return min(max_spreading, next_spread)

    def _continue_processing(self):
        if self.job.last_step_result is not None:
            return (not self.stop) and self.job.last_step_result
        else:
            return not self.stop

    def _check_queue_credentials(self):
        """
        Checks the security token expiration timestamp on the queue and
        refreshes it if necessary
        """
        if self.queue._credentials_expiring_soon():
            self.log.debug(
                f"refreshing security credentials for queue "
                f"`{self.queue.name}`"
            )
            self.queue = self.queue_adapter.reload(self.queue)

    def _run_functions(self):
        # Each worker has its own dedicated log
        self.job = QalxJob(self)
        _startup = True
        # avoid pickle errors
        self.steps = {}
        for f_name, f_dill in self._steps.items():  # de-serialise functions
            self.log.debug(f"try to undill step `{f_name}`")
            try:
                self.steps[f_name] = dill.loads(f_dill)
            except TypeError:
                self.steps[f_name] = f_dill

        self.log.debug("running")
        if self.steps.get("begin_function"):
            self.job._current_step = "Begin"

            self.steps.get("begin_function")(self.job)
            self._check_worker_signal()
            self._check_cold_exit(reload_signal=False)

        kill_after = self.config["KILL_AFTER"]
        spread_factor = 1
        no_message_attempts = 0
        # Checks the worker signal in case the bot is started in a "stopped"
        # state
        self._check_worker_signal()

        while True:
            # Check if the credentials need refreshing. They may have expired
            # due to `_wait_with_spreading` or for another machine related
            # issue (i.e. machine hibernate)
            self._check_queue_credentials()
            if self.stop:
                spread_factor = self._wait_with_spreading(spread_factor)
                self.log.debug(
                    f"Stop signal on Worker waited with factor {spread_factor}"
                )
            else:
                q_msgs = self.queue_adapter.get_messages(worker=self)
                self.log.debug(f"got {len(q_msgs)} messages")
                if q_msgs:
                    for q_msg in q_msgs:
                        if not _startup:
                            self.job = QalxJob(self)
                        _startup = False
                        self.job._queue_message = q_msg
                        self.log.debug(f"process {q_msg.body}")
                        self._process_functions()
                        self._check_worker_signal()
                    no_message_attempts = 0  # reset the attempt counter
                    spread_factor = 1  # reset any wait spreading
                else:
                    spread_factor = self._wait_with_spreading(spread_factor)
                    no_message_attempts += 1

            if (kill_after is not None) and (no_message_attempts >= kill_after):
                self._do_exit(**QalxSignal._config_exit_update())

    def _process_functions(self):
        def process_as_far_as_possible(functions):
            for fn, step_name in functions:
                if self._continue_processing():
                    self.job._current_step = step_name
                    try:
                        # Update bot and worker states here so that the state is correct
                        # after a `resume` signal is sent
                        self.worker_adapter.update_state(
                            self.api_entity,
                            ValidStates.ACTIVE,
                            bot_entity=self.bot_entity,
                        )
                        self.bot_adapter.update_state(
                            self.bot_entity, ValidStates.ACTIVE
                        )
                        fn(self.job)
                    except Exception:
                        # This needs to be a bare exception to handle
                        # something going wrong with the user defined functions
                        _, exc_value, exc_traceback = sys.exc_info()
                        # Send as a `QalxError` to our handler otherwise we
                        # risk consuming exceptions from the users application
                        exception_hook(
                            exc_type=QalxError,
                            exc_value=exc_value,
                            exc_traceback=exc_traceback,
                            extra={
                                "entity_guid": self.job.entity.guid,
                                "entity_type": self.job.entity.entity_type,
                            },
                            log=self.log,
                        )

                    # check if we need to cold exit after every function
                    self._check_cold_exit()
                    if self.config["SAVE_INTERMITTENT"]:
                        self.job.save_entity()
                else:
                    break
            # Processing discontinued due to stop signal, failed step
            # function or completing processing
            self.worker_adapter.update_state(
                self.api_entity,
                ValidStates.IDLE,
                bot_entity=self.bot_entity,
            )
            if ValidStates.ACTIVE.value not in [
                worker.state for worker in self.bot_entity.workers
            ]:
                self.bot_adapter.update_state(self.bot_entity, ValidStates.IDLE)
            if not self.config["SKIP_FINAL_SAVE"]:
                self.job.save_entity()
            # `_workflow` is executed only if the flag on the job entity is True
            if self.job.do_workflow is True:
                self._workflow()
            # We delete the message from the queue once we've processed it
            self._delete_message()
            return self.job

        # PreLoad
        if self.steps.get("preload_function"):
            self.job._current_step = "PreLoad"
            self.steps.get("preload_function")(self.job)
            self._check_cold_exit()
        if self._continue_processing():
            # Process steps
            steps = (
                "OnLoad",
                "PreProcess",
                "Process",
                "PreCompletion",
                "PostProcess",
            )
            process_functions = [
                (self.steps.get(f"{s.lower()}_function"), s)
                for s in steps
                if self.steps.get(f"{s.lower()}_function")
            ]

            entity_type = self.job._queue_message.body["entity_type"]
            entity_guid = self.job._queue_message.body["entity_guid"]
            if self.job.get_entity(entity_type, entity_guid):
                process_as_far_as_possible(process_functions)
        else:
            self._delete_message()

    def run(self):
        """
        Run our custom functions.  Catch any exception
        that may occur with it and pass that to our custom
        uncaught exception handler
        """
        # This is the earliest the log can be configured to avoid
        # pickling issues on windows.  Any logging before this needs
        # to be done using a QalxSession#log instance
        self.configure_logging()
        try:
            self._run_functions()
        except Exception:
            # This needs to be a bare exception to handle something going
            # wrong with the Worker
            _, exc_value, exc_traceback = sys.exc_info()
            # Send as a `QalxError` to our handler otherwise we
            # risk consuming exceptions from the users application
            exception_hook(
                exc_type=QalxError,
                exc_value=exc_value,
                exc_traceback=exc_traceback,
                log=self.log,
            )
